import random
from typing import Dict, Optional, Tuple

import torch

from diagnnose.activations.activation_reader import ActivationReader
from diagnnose.activations.selection_funcs import return_all, union
from diagnnose.corpus import Corpus
from diagnnose.corpus.create_labels import create_labels_from_corpus
from diagnnose.extract import simple_extract
from diagnnose.models import LanguageModel
from diagnnose.typedefs.activations import (
    ActivationIndex,
    ActivationName,
    ActivationNames,
    SelectionFunc,
)
from diagnnose.typedefs.probe import ControlTask, DataDict, DataSplit


class DataLoader:
    """Reads in pickled activations that have been extracted, and
    creates a train/test split of activations and labels.

    Train/test split can be created in multiple ways:
    1. Using the test activations from a different corpus
    2. Using the activations from the same corpus for both, but the
    train/test split is defined on 2 separate selection_funcs.
    3. Based on a random 90/10 split.

    Parameters
    ----------
    corpus : Corpus
        Corpus containing the labels for each sentence.
    activations_dir : str, optional
        Directory containing the extracted activations. If not provided,
        new activations will be extracted. If ``create_new_activations``
        is set to ``True``, the newly extracted activations will be
        stored in this directory.
    test_activations_dir : str, optional
        Directory containing the extracted test activations. If not
        provided the train activation set will be split and partially
        used as test set.
    test_corpus : Corpus, optional
        Corpus containing the test labels for each sentence. Must be
        provided if `test_activations_dir` is provided.
    train_selection_func : SelectFunc, optional
        Selection function that determines whether a corpus item should
        be taken into account for training. If not provided all
        extracted activations will be used and split into a random
        train/test split.
    test_selection_func : SelectFunc, optional
        Selection function that determines whether a corpus item should
        be taken into account for testing. If not provided all
        extracted activations will be used and split into a random
        train/test split.
    control_task: ControlTask, optional
        Control task function of Hewitt et al. (2019), mapping a corpus
        item to a random label.
    train_test_ratio : float, optional
        Ratio of the train/test split. If separate test
        activations are provided this split won't be used.
        Defaults to None, but must be provided if no test_selection_func
        is passed.
    create_new_activations : bool, optional
        Toggle to create new activations based on corpus and model.
        Overwrites existing activations that might be present in
        ``activations_dir``.
    """

    def __init__(
        self,
        corpus: Corpus,
        activations_dir: Optional[str] = None,
        model: Optional[LanguageModel] = None,
        activation_names: Optional[ActivationNames] = None,
        train_selection_func: Optional[SelectionFunc] = None,
        test_activations_dir: Optional[str] = None,
        test_corpus: Optional[Corpus] = None,
        test_selection_func: Optional[SelectionFunc] = None,
        control_task: Optional[ControlTask] = None,
        train_test_ratio: Optional[float] = None,
        create_new_activations: bool = False,
    ) -> None:
        assert (
            None not in [test_corpus, test_selection_func]
            or test_selection_func is not None
            or train_test_ratio is not None
        ), (
            "Test split must be provided by either passing a test_corpus, a test_selection_func, "
            "or a train_test_ratio. The docstring of DataLoader contains more precise information."
        )

        self.activation_names = activation_names
        self.create_new_activations = create_new_activations

        selection_func = (
            union((train_selection_func, test_selection_func))
            if test_selection_func is not None and test_corpus is None
            else train_selection_func
        )

        train_activation_reader = self._create_activation_reader(
            model, corpus, activations_dir, selection_func
        )

        self.train_ids, self.test_ids = self._train_test_ids(
            corpus,
            train_activation_reader.selection_func,
            train_selection_func,
            test_selection_func if test_corpus is None else None,
            train_test_ratio if test_corpus is None else None,
        )

        self.train_split = self._create_data_split(
            train_activation_reader,
            corpus,
            self.train_ids,
            control_task,
        )

        if test_corpus is not None:
            test_activation_reader = self._create_test_split(
                model,
                test_activations_dir,
                test_corpus,
                test_selection_func,
            )
        else:
            test_activation_reader = train_activation_reader
            test_corpus = corpus

        self.test_split = self._create_data_split(
            test_activation_reader,
            test_corpus,
            self.test_ids,
            control_task,
        )

        self.label_vocab: Dict[str, int] = corpus.fields[corpus.labels_column].vocab

        assert len(self.test_split.labels) > 0, (
            "DataSplit contains no test labels, check whether test_selection_func, "
            "test_activation_dir or train_test_ratio are set up correctly"
        )

    def _create_activation_reader(
        self,
        model: Optional[LanguageModel],
        corpus: Corpus,
        activations_dir: str,
        selection_func: Optional[SelectionFunc],
    ) -> ActivationReader:
        if activations_dir is None or self.create_new_activations:
            assert model is not None

            activation_reader, _ = simple_extract(
                model,
                corpus,
                self.activation_names,
                activations_dir=activations_dir,
                selection_func=selection_func or return_all,
            )
            activation_reader.cat_activations = True
        else:
            activation_reader = ActivationReader(activations_dir, cat_activations=True)

        return activation_reader

    @staticmethod
    def _create_data_split(
        activation_reader: ActivationReader,
        corpus: Corpus,
        label_ids: ActivationIndex,
        control_task: Optional[ControlTask],
    ) -> DataSplit:
        """Creates a labels Tensor and returns a DataSplit.

        The DataSplit contains an ActivationReader, the labels, and,
        optionally, the control labels.
        """
        labels = create_labels_from_corpus(
            corpus,
            selection_func=activation_reader.selection_func,
        )
        labels = labels[label_ids]

        if control_task is not None:
            control_labels = create_labels_from_corpus(
                corpus,
                selection_func=activation_reader.selection_func,
                control_task=control_task,
            )
            control_labels = control_labels[label_ids]
        else:
            control_labels = None

        return DataSplit(activation_reader, labels, control_labels)

    def _create_test_split(
        self,
        model: Optional[LanguageModel],
        test_activations_dir: Optional[str],
        test_corpus: Optional[Corpus],
        test_selection_func: Optional[SelectionFunc],
    ) -> ActivationReader:
        """Creates the DataSplit for the test items that are retrieved
        from a different Corpus than the train items.
        """
        test_activation_reader = self._create_activation_reader(
            model,
            test_corpus,
            test_activations_dir,
            test_selection_func,
        )

        self.test_ids, _ = self._train_test_ids(
            test_corpus,
            test_activation_reader.selection_func,
            test_selection_func,
            None,
            None,
        )

        return test_activation_reader

    def load(
        self,
        activation_name: ActivationName,
    ) -> DataDict:
        """Creates train/test data split of activations

        Parameters
        ----------
        activation_name : ActivationName
            (layer, name) tuple indicating the activations to be read in

        Returns
        -------
        data_dict : DataDict
            Dictionary containing train and test activations, and their
            corresponding labels and, optionally, control labels.
        """
        train_activations = self.train_split.activation_reader[:, activation_name]
        test_activations = self.test_split.activation_reader[:, activation_name]

        train_activations = train_activations[self.train_ids]
        test_activations = test_activations[self.test_ids]

        train_labels = self.train_split.labels
        test_labels = self.test_split.labels

        train_control_labels = self.train_split.control_labels
        test_control_labels = self.test_split.control_labels

        data_dict = DataDict(
            train_activations,
            train_labels,
            train_control_labels,
            test_activations,
            test_labels,
            test_control_labels,
        )

        return data_dict

    @staticmethod
    def _train_test_ids(
        corpus: Corpus,
        orig_selection_func: SelectionFunc,
        train_selection_func: Optional[SelectionFunc],
        test_selection_func: Optional[SelectionFunc],
        train_test_ratio: Optional[float],
    ) -> Tuple[ActivationIndex, ActivationIndex]:
        """Creates a tensor mask for the train/test split.

        Mask is created based on the provided selection functions.

        Note that the train_selection_func takes precedence over the
        test_selection_func.

        The mask also takes the original selection_func into account,
        and will skip items that are not part of either the provided
        selection_funcs.
        """
        train_ids = []
        test_ids = []

        for item in corpus.examples:
            sen = getattr(item, corpus.sen_column)
            for pos in range(len(sen)):
                if orig_selection_func(pos, item):
                    if train_selection_func is None or train_selection_func(pos, item):
                        in_train = (
                            random.random() < train_test_ratio
                            if train_test_ratio is not None
                            else True
                        )
                        train_ids.append(in_train)
                        test_ids.append(not in_train)
                    elif test_selection_func and test_selection_func(pos, item):
                        train_ids.append(False)
                        test_ids.append(True)
                    else:
                        train_ids.append(False)
                        test_ids.append(False)

        return (
            torch.tensor(train_ids).to(torch.bool),
            torch.tensor(test_ids).to(torch.bool),
        )



from typing import Dict, List, Optional

from torchtext.data import Example

from diagnnose.corpus import Corpus
from diagnnose.typedefs.syntax import SyntaxEvalCorpora

from ..task import SyntaxEvalTask
from .warstadt_preproc import ENVS, create_downstream_corpus, preproc_warstadt


class WarstadtTask(SyntaxEvalTask):
    def initialize(
        self, path: str, subtasks: Optional[List[str]] = None
    ) -> SyntaxEvalCorpora:
        """Performs the initialization for the tasks of
        Marvin & Linzen (2018)

        Arxiv link: https://arxiv.org/pdf/1808.09031.pdf

        Repo: https://github.com/BeckyMarvin/LM_syneval

        Parameters
        ----------
        path : str
            Path to directory containing the Marvin datasets that can be
            found in the github repo.
        subtasks : List[str], optional
            The downstream tasks that will be tested. If not provided this
            will default to the full set of conditions.

        Returns
        -------
        corpora : Dict[str, Corpus]
            Dictionary mapping a subtask to a Corpus.
        """
        subtasks: List[str] = subtasks or ENVS

        corpora: SyntaxEvalCorpora = {}

        orig_corpus = preproc_warstadt(path)

        for env in subtasks:
            raw_corpus = create_downstream_corpus(orig_corpus, envs=[env])

            header = raw_corpus[0].split("\t")
            tokenize_columns = ["sen", "counter_sen"]
            fields = Corpus.create_fields(
                header, tokenize_columns=tokenize_columns, tokenizer=self.tokenizer
            )
            examples = [
                Example.fromlist(line.split("\t"), fields) for line in raw_corpus[1:]
            ]
            corpus = Corpus(examples, fields)

            corpora[env] = corpus

        return corpora



import abc
from typing import TYPE_CHECKING, Dict, List, Optional, Type

import torch
from torch import Tensor
from transformers import BatchEncoding

from .gcd_tensor import GCDTensor
from .shapley_tensor import ShapleyTensor

if TYPE_CHECKING:
    from diagnnose.models import LanguageModel


tensor_types: Dict[str, Type[ShapleyTensor]] = {
    "ShapleyTensor": ShapleyTensor,
    "GCDTensor": GCDTensor,
}


class Decomposer(abc.ABC):
    """Abstract base class for Decomposer classes.

    A Decomposer takes care of dividing the input features into the
    desired partition of contributions.
    """

    def __init__(
        self,
        model: "LanguageModel",
        num_samples: Optional[int] = None,
        tensor_type: str = "ShapleyTensor",
    ):
        self.model = model
        self.num_samples = num_samples
        self.tensor_type = tensor_types[tensor_type]

    @abc.abstractmethod
    def decompose(self, batch_encoding: BatchEncoding) -> ShapleyTensor:
        raise NotImplementedError

    @abc.abstractmethod
    def wrap_inputs_embeds(self, input_ids: Tensor) -> ShapleyTensor:
        raise NotImplementedError


class ShapleyDecomposer(Decomposer):
    """A ShapleyDecomposer propagates all input feature contributions
    simultaneously.

    That is, an input sequence of :math:`n` features will be transformed
    into a ShapleyTensor containing :math:`n` feature contributions.

    Concretely: if we have an input tensor :math:`X` of shape:
    ``(num_features, input_dim)`` we can express this as a sum of
    features:
    :math:`X = \\sum_i^n \\phi^i`, where :math:`\\phi^i` is also of
    shape ``(num_features, input_dim)``, with
    :math:`\\phi^i_j =
    \\begin{cases}X_j&i=j\\\\0&\\textit{otherwise}\\end{cases}`

    Without approximations this way of partitioning scales
    exponentially in the number of input features, quickly becoming
    infeasible when :math:`n > 10`.
    """

    def decompose(self, batch_encoding: BatchEncoding) -> ShapleyTensor:
        input_ids = torch.tensor(batch_encoding["input_ids"])
        inputs_embeds = self.wrap_inputs_embeds(input_ids)

        with torch.no_grad():
            shapley_out = self.model(
                inputs_embeds=inputs_embeds,
                input_lengths=batch_encoding.data.get("length", None),
                compute_out=True,
                only_return_top_embs=True,
            )

        return shapley_out

    def wrap_inputs_embeds(self, input_ids: Tensor) -> ShapleyTensor:
        # Shape: batch_size x max_sen_len x nhid
        inputs_embeds = self.model.create_inputs_embeds(input_ids)

        # First contribution corresponds to contributions stemming from bias terms within the
        # model itself.
        contributions = [torch.zeros_like(inputs_embeds)]

        # Each individual contribution is set to its corresponding input feature, and set to
        # zero on all other positions.
        for w_idx in range(inputs_embeds.shape[1]):
            contribution = torch.zeros_like(inputs_embeds)
            contribution[:, w_idx] = inputs_embeds[:, w_idx]
            contributions.append(contribution)

        shapley_in = self.tensor_type(
            inputs_embeds,
            contributions=contributions,
            validate=True,
            num_samples=self.num_samples,
        )

        return shapley_in


class ContextualDecomposer(Decomposer):
    """A ContextualDecomposer propagates each input feature
    contribution individually, set out against the contributions of all
    other features combined.

    This idea has been proposed in Murdocht et al., (2018):
    https://arxiv.org/abs/1801.05453

    An input sequence of :math:`n` features will be transformed
    into a ShapleyTensor containing :math:`2` feature contributions:
    one containing the contributions of the feature of interest
    (:math:`\\beta`), and one containing the contributions of all
    other features (:math:`\\gamma`).

    Concretely: if we have an input tensor :math:`X` of shape:
    ``(num_features, input_dim)`` we can express this as a sum of
    features:
    :math:`X = \\beta^i + \\gamma^i`, where both :math:`\\beta` and
    :math:`\\gamma` are also of shape ``(num_features, input_dim)``,
    with :math:`\\beta^i_j =
    \\begin{cases}X_j&i=j\\\\0&\\textit{otherwise}\\end{cases}` and
    :math:`\\gamma^i_j =
    \\begin{cases}X_j&i\\neq j\\\\0&\\textit{otherwise}\\end{cases}`

    This way of partitioning scales polynomially in the number of input
    features, but requires a separate forward pass for each individual
    feature contribution :math:`\\beta^i`.
    """

    def decompose(self, batch_encoding: BatchEncoding) -> ShapleyTensor:
        input_ids = torch.tensor(batch_encoding["input_ids"])
        shapley_tensors = self.wrap_inputs_embeds(input_ids)

        all_contributions = []

        for w_idx, inputs_embeds in enumerate(shapley_tensors):
            with torch.no_grad():
                out, (beta, _gamma) = self.model(
                    inputs_embeds=inputs_embeds,
                    input_lengths=batch_encoding.data.get("length", None),
                    compute_out=True,
                    only_return_top_embs=True,
                )
            all_contributions.append(beta)

        return GCDTensor(out, all_contributions)

    def wrap_inputs_embeds(self, input_ids: Tensor) -> List[ShapleyTensor]:
        inputs_embeds = self.model.create_inputs_embeds(input_ids)

        all_shapley_in = [
            GCDTensor(
                inputs_embeds,
                contributions=[torch.zeros_like(inputs_embeds), inputs_embeds],
                validate=True,
                num_samples=self.num_samples,
            )
        ]

        for w_idx in range(inputs_embeds.shape[1]):
            beta = torch.zeros_like(inputs_embeds)
            gamma = inputs_embeds.clone()

            beta[:, w_idx] = gamma[:, w_idx]
            gamma[:, w_idx] = 0.0

            contributions = [beta, gamma]

            shapley_in = GCDTensor(
                inputs_embeds,
                contributions=contributions,
                validate=False,
                num_samples=self.num_samples,
                baseline_partition=1,
            )

            all_shapley_in.append(shapley_in)

        return all_shapley_in



from typing import Any, Dict, Set

from unidecode import unidecode

from .w2i import W2I


class C2I(W2I):
    """Vocabulary containing character-level information.

    Adapted from: https://github.com/tensorflow/models/tree/master/research/lm_1b
    """

    def __init__(
        self, w2i: Dict[str, int], max_word_length: int = 50, **kwargs: Any
    ) -> None:
        super().__init__(w2i, **kwargs)
        self._max_word_length = max_word_length
        chars_set: Set[str] = set()

        for word in w2i:
            chars_set |= set(word)

        free_ids = []
        for i in range(256):
            if chr(i) in chars_set:
                continue
            free_ids.append(chr(i))

        if len(free_ids) < 5:
            raise ValueError("Not enough free char ids: %d" % len(free_ids))

        self.eos_char = free_ids[1]  # <end sentence>
        self.bow_char = free_ids[2]  # <begin word>
        self.eow_char = free_ids[3]  # <end word>
        self.pad_char = free_ids[4]  # <padding>

        self._word_char_ids = {}

        for w in self.w2i.keys():
            self._word_char_ids[w] = self._convert_word_to_char_ids(w)

    @property
    def max_word_length(self) -> int:
        return self._max_word_length

    def _convert_word_to_char_ids(self, word: str):
        import numpy as np

        code = np.zeros([self.max_word_length], dtype=np.int32)
        code[:] = ord(self.pad_char)

        if len(word) > self.max_word_length - 2:
            word = word[: self.max_word_length - 2]
        cur_word = self.bow_char + word + self.eow_char
        for j in range(len(cur_word)):
            code[j] = ord(cur_word[j])
        return code.reshape((1, 1, -1))

    def token_to_char_ids(self, token: str):
        if not all(ord(c) < 256 for c in token):
            token = unidecode(token)

        if token in self._word_char_ids:
            char_ids = self._word_char_ids[token]
        else:
            char_ids = self._convert_word_to_char_ids(token)
            self._word_char_ids[token] = char_ids

        return char_ids



from typing import Optional

import torch
from torch import Tensor

from diagnnose.corpus import Corpus
from diagnnose.typedefs.activations import SelectionFunc
from diagnnose.typedefs.probe import ControlTask


def create_labels_from_corpus(
    corpus: Corpus,
    selection_func: SelectionFunc = lambda sen_id, pos, example: True,
    control_task: Optional[ControlTask] = None,
) -> Tensor:
    """Creates labels based on the selection_func that was used during
    extraction.

    Parameters
    ----------
    corpus : Corpus
        Labeled corpus containing sentence and label information.
    selection_func: SelectFunc, optional
        Function that determines whether a label should be stored.
    control_task: ControlTask, optional
        Control task function of Hewitt et al. (2019), mapping a corpus
        item to a random label.
    """
    all_labels = []

    for item in corpus.examples:
        label_idx = 0

        sen = getattr(item, corpus.sen_column)
        labels = getattr(item, corpus.labels_column)
        if isinstance(labels, str):
            labels = labels.split()

        each_token_labeled = len(sen) == len(labels)

        for wpos in range(len(sen)):
            if selection_func(wpos, item):
                if control_task is not None:
                    label = control_task(wpos, item)
                else:
                    label = labels[label_idx]

                all_labels.append(label)

                if not each_token_labeled and len(labels) > 1:
                    label_idx += 1
            if each_token_labeled:
                label_idx += 1

    # Create new label vocab that only contains the labels that have been selected
    label_vocab = {label: idx for idx, label in enumerate(set(all_labels))}
    corpus.fields[corpus.labels_column].vocab = label_vocab

    return torch.tensor([label_vocab[label] for label in all_labels])



import os
import warnings
from time import time
from typing import Any, Dict, Optional

import torch
from torch import Tensor

from diagnnose.typedefs.activations import ActivationName
from diagnnose.typedefs.probe import DataDict, DCConfig
from diagnnose.utils.pickle import dump_pickle

from .data_loader import DataLoader
from .logreg import L1NeuralNetClassifier, LogRegModule


class DCTrainer:
    """Trains Diagnostic Classifiers (DC) on extracted activation data.

    For each activation that is part of the provided activation_names
    argument a different classifier will be trained.

    Parameters
    ----------
    data_loader : DataLoader
        ``DataLoader`` that contains the activations and labels on
        which the DCs will be trained. A ``DataLoader`` can contain
        activations for multiple layers and gates of a model, for which
        separate DCs will be trained and evaluated.
    save_dir : str
        Directory to which trained models will be saved.
    lr : float, optional
        Learning rate of the linear classifier that is used during
        training. Defaults to 0.01.
    max_epochs : int, optional
        Maximum number of training epochs used for cross-validation.
        Defaults to 10.
    rank : int, optional
        Matrix rank of the linear classifier. Defaults to the full rank
        if not provided.
    lambda1 : float, optional
        Coefficient for L1 regularization that can be increased to
        induce sparsity in the diagnostic classifier. Defaults to 0.,
        indicating no L1 regularization.
    verbose : int, optional
        Set to any positive number for verbosity. Defaults to 0.

    Attributes
    ----------
    classifier : Classifier
        Current classifier that is being trained.
    """

    def __init__(
        self,
        data_loader: DataLoader,
        save_dir: str,
        lr: float = 0.01,
        max_epochs: int = 10,
        rank: Optional[int] = None,
        lambda1: float = 0.0,
        verbose: int = 0,
    ) -> None:
        self.save_dir = save_dir
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        self.classifier = None

        self.data_loader = data_loader

        self.dc_config = DCConfig(lr, max_epochs, rank, lambda1, verbose)

    def train(self) -> Dict[ActivationName, Any]:
        """Trains DCs on multiple activation names."""

        full_results_dict = {}

        for activation_name in self.data_loader.activation_names:
            results_dict = self._train_one_dc(activation_name)
            full_results_dict[activation_name] = results_dict

        return full_results_dict

    def _train_one_dc(self, activation_name: ActivationName) -> Dict[str, Any]:
        """ Initiates training the DC on 1 activation type. """
        data_dict: DataDict = self.data_loader.load(activation_name)

        if self.dc_config.verbose > 0:
            train_size = data_dict.train_activations.size(0)
            test_size = data_dict.test_activations.size(0)
            print(f"train/test: {train_size}/{test_size}")
            print(f"\nStarting fitting model on {activation_name}...")

        # Train
        self._fit(data_dict.train_activations, data_dict.train_labels)

        results_dict = self._eval(data_dict.test_activations, data_dict.test_labels)

        self._save_classifier(activation_name)

        if data_dict.train_control_labels is not None:
            self._fit(data_dict.train_activations, data_dict.train_control_labels)
            control_results = self._eval(
                data_dict.test_activations, data_dict.test_control_labels
            )
            results_dict["control"] = control_results
            results_dict["selectivity"] = (
                results_dict["accuracy"] - control_results["accuracy"]
            )
            self._save_classifier(activation_name, postfix="_control")

        self._save_results(results_dict, activation_name)

        return results_dict

    def _fit(self, activations: Tensor, labels: Tensor) -> None:
        self._reset_classifier(activations.size(1), len(torch.unique(labels)))

        start_time = time()

        self.classifier.fit(activations, labels)

        if self.dc_config.verbose > 0:
            print(f"Fitting done in {time() - start_time:.2f}s")

    def _eval(self, activations: Tensor, labels: Tensor) -> Dict[str, Any]:
        try:
            import sklearn.metrics as metrics
        except ImportError:
            warnings.warn("sklearn.metrics is needed for DC evaluation")
            raise

        pred_y = self.classifier.predict(activations)

        acc = metrics.accuracy_score(labels, pred_y)
        f1 = metrics.f1_score(labels, pred_y, average="micro")
        mcc = metrics.matthews_corrcoef(labels, pred_y)
        cm = metrics.confusion_matrix(labels, pred_y)

        results_dict = {"accuracy": acc, "f1": f1, "mcc": mcc, "confusion_matrix": cm}

        return results_dict

    def _save_classifier(self, activation_name: ActivationName, postfix: str = ""):
        if self.save_dir is not None:
            l, name = activation_name
            fn = f"{name}_l{l}" + postfix
            model_path = os.path.join(self.save_dir, fn + ".pt")
            torch.save(self.classifier.module.state_dict(), model_path)

    def _save_results(
        self, results_dict: Dict[str, Any], activation_name: ActivationName
    ) -> None:
        if self.dc_config.verbose > 0:
            for k, v in results_dict.items():
                print(k, v, "", sep="\n")
            print("Label vocab:", self.data_loader.label_vocab)

        if self.save_dir is not None:
            l, name = activation_name
            results_path = os.path.join(self.save_dir, f"{name}_l{l}_results.pickle")
            dump_pickle(results_dict, results_path)

    def _reset_classifier(self, ninp: int, nout: int) -> None:
        dc_config_dict = self.dc_config._asdict()
        rank = dc_config_dict.pop("rank")

        self.classifier = L1NeuralNetClassifier(
            LogRegModule(ninp=ninp, nout=nout, rank=rank),
            optimizer=torch.optim.Adam,
            **dc_config_dict,
        )



import os
import sys
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
from torch import Tensor

from diagnnose.models.recurrent_lm import RecurrentLM
from diagnnose.tokenizer import create_char_vocab
from diagnnose.tokenizer.c2i import C2I
from diagnnose.typedefs.activations import ActivationDict, LayeredTensors


class GoogleLM(RecurrentLM):
    """Reimplementation of the LM of Jozefowicz et al. (2016).

    Paper: https://arxiv.org/abs/1602.02410
    Lib: https://github.com/tensorflow/models/tree/master/research/lm_1b

    This implementation allows for only a subset of the SoftMax to be
    loaded in, to alleviate RAM usage.

    Parameters
    ----------
    pbtxt_path : str
        Path to the .pbtxt file containing the GraphDef model setup.
    ckpt_dir : str
        Path to folder containing parameter checkpoint files.
    full_vocab_path : str
        Path to the full model vocabulary of 800k tokens. Note that
        `vocab_path` can be passed along as well, pointing toward the
        corpus that will be extracted. In that case only a subset of
        the model softmax will be loaded in.
    corpus_vocab_path : str, optional
        Path to the corpus for which a vocabulary will be created. This
        allows for only a subset of the model softmax to be loaded in.
    create_decoder : bool
        Toggle to load in the (partial) softmax weights. Can be set to
        false in case no decoding projection needs to be made, as is
        the case during activation extraction, for example.
    """

    forget_offset = 1
    ih_concat_order = ["i", "h"]
    sizes = {l: {"x": 1024, "h": 1024, "c": 8192} for l in range(2)}
    split_order = ["i", "g", "f", "o"]
    use_char_embs = True

    def __init__(
        self,
        pbtxt_path: str,
        ckpt_dir: str,
        full_vocab_path: str,
        corpus_vocab_path: Optional[Union[str, List[str]]] = None,
        create_decoder: bool = True,
        device: str = "cpu",
    ) -> None:
        super().__init__()
        print("Loading pretrained model...")

        vocab: C2I = create_char_vocab(corpus_vocab_path or full_vocab_path)

        self.device = device
        self.decoder_w = None
        self.decoder_b = None

        try:
            self.encoder = CharCNN(pbtxt_path, ckpt_dir, vocab, device)
            self.lstm = LSTM(
                ckpt_dir, self.num_layers, self.split_order, self.forget_offset, device
            )
            if create_decoder:
                self.decoder = SoftMax(
                    vocab, full_vocab_path, ckpt_dir, self.sizes[1]["h"], device
                )
                self.decoder_w = self.decoder.decoder_w
                self.decoder_b = self.decoder.decoder_b
        except ImportError:
            raise ImportError("tensorflow and protobuf are needed for GoogleLM")

        print("Model initialisation finished.")

    @property
    def vocab(self) -> C2I:
        return self.encoder.vocab

    @property
    def weight(self) -> LayeredTensors:
        return self.lstm.weight

    @property
    def bias(self) -> LayeredTensors:
        return self.lstm.bias

    @property
    def peepholes(self) -> ActivationDict:
        return self.lstm.peepholes

    def forward(
        self,
        tokens: List[str],
        prev_activations: Optional[ActivationDict] = None,
        compute_out: bool = True,
    ) -> Tuple[Optional[Tensor], ActivationDict]:
        # Create the embeddings of the input words
        embs = self.encoder(tokens)

        if prev_activations is None:
            prev_activations = self.init_hidden(1)

        logits, activations = self.lstm(embs, prev_activations)

        if compute_out and self.decoder_w is not None:
            out = self.decoder_w @ logits + self.decoder_b
        else:
            out = None

        return out, activations


class CharCNN(nn.Module):
    def __init__(self, pbtxt_path: str, ckpt_dir: str, vocab: C2I, device: str) -> None:
        print("Loading CharCNN...")
        super().__init__()

        self.cnn_sess, self.cnn_t = self._load_char_cnn(pbtxt_path, ckpt_dir)
        self.cnn_embs: Dict[str, Tensor] = {}
        self.vocab = vocab
        self.device = device

    @staticmethod
    def _load_char_cnn(pbtxt_path: str, ckpt_dir: str) -> Any:
        import tensorflow as tf
        from google.protobuf import text_format

        ckpt_file = os.path.join(ckpt_dir, "ckpt-char-embedding")

        with tf.Graph().as_default():
            sys.stderr.write("Recovering graph.\n")
            with tf.gfile.FastGFile(pbtxt_path, "r") as f:
                s = f.read()
                gd = tf.GraphDef()
                text_format.Merge(s, gd)

            t = dict()
            [t["char_inputs_in"], t["all_embs"]] = tf.import_graph_def(
                gd, {}, ["char_inputs_in:0", "all_embs_out:0"], name=""
            )

            sess = tf.Session(config=tf.ConfigProto(allow_soft_placement=True))
            sess.run(f"save/Assign", {"save/Const:0": ckpt_file})
            # The following was recovered from the graph structure, the first 62 assign modules
            # relate to the parameters of the char CNN.
            for i in range(1, 62):
                sess.run(f"save/Assign_{i}", {"save/Const:0": ckpt_file})

        return sess, t

    def forward(self, tokens: List[str]) -> Tensor:
        embs = torch.zeros((len(tokens), 1024), device=self.device)
        for i, token in enumerate(tokens):
            if token not in self.cnn_embs:
                char_ids = self.vocab.token_to_char_ids(token)
                input_dict = {self.cnn_t["char_inputs_in"]: char_ids}
                emb = torch.from_numpy(
                    self.cnn_sess.run(self.cnn_t["all_embs"], input_dict)
                ).to(self.device)
                self.cnn_embs[token] = emb
            else:
                emb = self.cnn_embs[token]
            embs[i] = emb

        return embs


class LSTM(nn.Module):
    def __init__(
        self,
        ckpt_dir: str,
        num_layers: int,
        split_order: List[str],
        forget_offset: int,
        device: str,
    ) -> None:
        super().__init__()

        print("Loading LSTM...")

        self.num_layers = num_layers
        self.split_order = split_order
        self.forget_offset = forget_offset

        # Projects hidden+input (2*1024) onto cell state dimension (8192)
        self.weight: LayeredTensors = {}
        self.bias: LayeredTensors = {}

        # Projects cell state dimension (8192) back to hidden dimension (1024)
        self.weight_P: LayeredTensors = {}
        # The 3 peepholes are weighted by a diagonal matrix
        self.peepholes: ActivationDict = {}

        self._load_lstm(ckpt_dir, device)

    def _load_lstm(self, ckpt_dir: str, device: str) -> None:
        from tensorflow.python.pywrap_tensorflow import NewCheckpointReader

        lstm_reader = NewCheckpointReader(os.path.join(ckpt_dir, "ckpt-lstm"))

        for l in range(self.num_layers):
            # Model weights are divided into 8 chunks
            # Shape: (2048, 32768)
            self.weight[l] = torch.cat(
                [
                    torch.from_numpy(lstm_reader.get_tensor(f"lstm/lstm_{l}/W_{i}"))
                    for i in range(8)
                ],
                dim=0,
            )

            # Shape: (32768,)
            self.bias[l] = torch.from_numpy(lstm_reader.get_tensor(f"lstm/lstm_{l}/B"))

            # Shape: (8192, 1024)
            self.weight_P[l] = torch.cat(
                [
                    torch.from_numpy(lstm_reader.get_tensor(f"lstm/lstm_{l}/W_P_{i}"))
                    for i in range(8)
                ],
                dim=0,
            )

            for p in ["f", "i", "o"]:
                # Shape: (8192, 8192)
                self.peepholes[l, p] = torch.from_numpy(
                    lstm_reader.get_tensor(f"lstm/lstm_{l}/W_{p.upper()}_diag")
                )

            # Cast to float32 tensors
            self.weight[l] = self.weight[l].to(device)
            self.weight_P[l] = self.weight_P[l].to(device)
            self.bias[l] = self.bias[l].to(device)
            for p in ["f", "i", "o"]:
                self.peepholes[l, p] = self.peepholes[l, p].to(device)

    def forward_step(
        self, layer: int, emb: Tensor, prev_hx: Tensor, prev_cx: Tensor
    ) -> ActivationDict:
        proj: Tensor = torch.cat((emb, prev_hx), dim=1) @ self.weight[layer]
        proj += self.bias[layer]

        split_proj: Dict[str, Tensor] = dict(
            zip(self.split_order, torch.chunk(proj, 4, dim=1))
        )

        f_g = torch.sigmoid(
            split_proj["f"] + prev_cx * self.peepholes[layer, "f"] + self.forget_offset
        )
        i_g = torch.sigmoid(split_proj["i"] + prev_cx * self.peepholes[layer, "i"])
        c_tilde_g = torch.tanh(split_proj["g"])

        cx = f_g * prev_cx + i_g * c_tilde_g
        o_g = torch.sigmoid(split_proj["o"] + cx * self.peepholes[layer, "o"])
        hx = (o_g * torch.tanh(cx)) @ self.weight_P[layer]

        return {
            (layer, "emb"): emb,
            (layer, "hx"): hx,
            (layer, "cx"): cx,
            (layer, "f_g"): f_g,
            (layer, "i_g"): i_g,
            (layer, "o_g"): o_g,
            (layer, "c_tilde_g"): c_tilde_g,
        }

    def forward(
        self, input_: Tensor, prev_activations: ActivationDict
    ) -> Tuple[Optional[Tensor], ActivationDict]:
        # Iteratively compute and store intermediate rnn activations
        activations: ActivationDict = {}

        for l in range(self.num_layers):
            prev_hx = prev_activations[l, "hx"]
            prev_cx = prev_activations[l, "cx"]
            activations.update(self.forward_step(l, input_, prev_hx, prev_cx))
            input_ = activations[l, "hx"]

        return input_, activations


class SoftMax:
    def __init__(
        self,
        vocab: C2I,
        full_vocab_path: str,
        ckpt_dir: str,
        hidden_size_h: int,
        device: str,
    ) -> None:
        print("Loading SoftMax...")
        self.decoder_w: Tensor = torch.zeros((len(vocab), hidden_size_h), device=device)
        self.decoder_b: Tensor = torch.zeros(len(vocab), device=device)

        self._load_softmax(vocab, full_vocab_path, ckpt_dir)

    def _load_softmax(self, vocab: C2I, full_vocab_path: str, ckpt_dir: str) -> None:
        from tensorflow.python.pywrap_tensorflow import NewCheckpointReader

        with open(full_vocab_path) as f:
            full_vocab: List[str] = f.read().strip().split("\n")

        bias_reader = NewCheckpointReader(os.path.join(ckpt_dir, "ckpt-softmax8"))
        full_bias = torch.from_numpy(bias_reader.get_tensor("softmax/b"))

        # SoftMax is chunked into 8 arrays of size 100000x1024
        for i in range(8):
            sm_reader = NewCheckpointReader(os.path.join(ckpt_dir, f"ckpt-softmax{i}"))

            sm_chunk = torch.from_numpy(sm_reader.get_tensor(f"softmax/W_{i}"))
            bias_chunk = full_bias[i : full_bias.size(0) : 8]
            vocab_chunk = full_vocab[i : full_bias.size(0) : 8]

            for j, w in enumerate(vocab_chunk):
                sm = sm_chunk[j]
                bias = bias_chunk[j]

                if w in vocab:
                    self.decoder_w[vocab[w]] = sm
                    self.decoder_b[vocab[w]] = bias

                if w == "</S>":
                    self.decoder_w[vocab[vocab.eos_token]] = sm
                    self.decoder_b[vocab[vocab.eos_token]] = bias
                elif w == "<UNK>":
                    self.decoder_w[vocab[vocab.unk_token]] = sm
                    self.decoder_b[vocab[vocab.unk_token]] = bias



import glob
import os
import warnings
from typing import Any, Dict, List, Optional

import torch
from torch import Tensor
from transformers import PreTrainedTokenizer

from diagnnose.activations.selection_funcs import final_token, only_mask_token
from diagnnose.corpus import Corpus
from diagnnose.extract import simple_extract
from diagnnose.models import LanguageModel
from diagnnose.typedefs.activations import SelectionFunc
from diagnnose.typedefs.syntax import ResultsDict, SyntaxEvalCorpora


class SyntaxEvalTask:
    """Base class for syntactic evaluation tasks, from which specific
    tasks can inherit.

    Parameters
    ----------
    model : LanguageModel
        Language model for which the accuracy is calculated.
    tokenizer : PreTrainedTokenizer
        The model tokenizer that converts tokens into indices.
    config : Dict[str, Any]
        Configuration dictionary containing the setup for task
        initialization.
    use_full_model_probs : bool, optional
        Toggle to calculate the full model probs for the NPI
        sentences. If set to False only the NPI logits will be
        compared, instead of their Softmax probabilities. Defaults
        to True.
    ignore_unk : bool, optional
        Ignore cases for which at least one of the cases of the verb
        is not part of the model's tokenizer. Defaults to False.
    """

    def __init__(
        self,
        model: LanguageModel,
        tokenizer: PreTrainedTokenizer,
        ignore_unk: bool,
        use_full_model_probs: bool,
        **config: Dict[str, Any],
    ):
        model.eval()
        self.model = model
        self.tokenizer = tokenizer

        self.ignore_unk = ignore_unk
        self.use_full_model_probs = use_full_model_probs

        # If a single subtask is passed as cmd arg it is not converted to a list yet
        if isinstance(config.get("subtasks", None), str):
            config["subtasks"] = [config["subtasks"]]

        self.corpora: SyntaxEvalCorpora = self.initialize(**config)

    def initialize(
        self, path: str, header: Optional[List[str]] = None
    ) -> SyntaxEvalCorpora:
        if header is None:
            header = ["sen", "token", "counter_token"]

        assert "sen" in header
        assert "token" in header
        assert "counter_sen" in header or "counter_token" in header

        corpora = {}

        if os.path.isdir(path):
            for file in glob.glob(os.path.join(path, "*")):
                corpus = Corpus.create(file, header=header, tokenizer=self.tokenizer)
                task_name = file.split("/")[-1].split(".")[0]
                corpora[task_name] = corpus
        elif os.path.isfile(path):
            corpus = Corpus.create(path, header=header, tokenizer=self.tokenizer)
            task_name = path.split("/")[-1].split(".")[0]
            corpora[task_name] = corpus
        else:
            raise FileNotFoundError("Path to task is not found")

        return corpora

    def run(self) -> ResultsDict:
        """Performs the syntactic evaluation task that is initialised.

        Returns
        -------
        results : ResultsDict
            Dictionary mapping a task to a task condition to the model
            accuracy.
        """
        results: ResultsDict = {}

        for subtask, subtask_corpora in self.corpora.items():
            if isinstance(subtask_corpora, Corpus):
                accuracy = self._run_corpus(subtask_corpora)
                results[subtask] = accuracy
            else:
                for condition, corpus in subtask_corpora.items():
                    accuracy = self._run_corpus(corpus)

                    results.setdefault(subtask, {})[condition] = accuracy

        return results

    def _run_corpus(self, corpus: Corpus) -> float:
        if self.model.is_recurrent:
            selection_func = final_token("sen")
        else:
            selection_func = only_mask_token(self.tokenizer.mask_token, "sen")

        activations = self._calc_final_hidden(corpus, selection_func)

        if "counter_sen" in corpus.fields:
            if self.model.is_recurrent:
                selection_func = final_token("counter_sen")
            else:
                selection_func = only_mask_token(
                    self.tokenizer.mask_token, "counter_sen"
                )

            corpus.sen_column = "counter_sen"
            counter_activations = self._calc_final_hidden(corpus, selection_func)
        else:
            counter_activations = None

        accuracy = self._calc_accuracy(
            corpus,
            activations,
            counter_activations=counter_activations,
        )

        return accuracy

    def _calc_final_hidden(
        self,
        corpus: Corpus,
        selection_func: SelectionFunc,
    ) -> Tensor:
        activation_name = (self.model.top_layer, "hx")

        activation_reader, _ = simple_extract(
            self.model,
            corpus,
            [activation_name],
            batch_size=len(corpus),
            selection_func=selection_func,
        )

        activations = activation_reader.activation_dict[activation_name]

        return activations

    def _calc_accuracy(
        self,
        corpus: Corpus,
        activations: Tensor,
        counter_activations: Optional[Tensor] = None,
    ) -> float:
        mask = self._create_unk_sen_mask(corpus)

        activations = activations[mask]

        token_ids = torch.tensor(
            [self.tokenizer.convert_tokens_to_ids(ex.token) for ex in corpus]
        )
        token_ids = token_ids[mask]

        if counter_activations is None:
            counter_token_ids = torch.tensor(
                [
                    self.tokenizer.convert_tokens_to_ids(ex.counter_token)
                    for ex in corpus
                ]
            )
            counter_token_ids = counter_token_ids[mask]

            accuracy = self._single_context_accuracy(
                activations, token_ids, counter_token_ids
            )
        else:
            accuracy = self._dual_context_accuracy(
                activations, counter_activations[mask], token_ids
            )

        return accuracy

    def _create_unk_sen_mask(self, corpus: Corpus) -> Tensor:
        """
        Creates a tensor mask for sentences that contain at least one
        token that is not part of the model's tokenizer.
        """
        mask = torch.ones(len(corpus), dtype=torch.bool)
        if not self.ignore_unk:
            return mask

        for idx, ex in enumerate(corpus):
            for w in ex.sen:
                if w not in self.tokenizer.vocab:
                    mask[idx] = False
                    warnings.warn(f"'{w}' is not part of model's tokenizer!")

        return mask

    def _single_context_accuracy(
        self, activations: Tensor, token_ids: Tensor, counter_token_ids: Tensor
    ) -> float:
        """ Computes accuracy for comparing P(w1|h) > P(w2|h). """
        logits = self._decode(activations, token_ids)
        counter_logits = self._decode(activations, counter_token_ids)

        return torch.mean((logits >= counter_logits).to(torch.float)).item()

    def _dual_context_accuracy(
        self,
        activations: Tensor,
        counter_activations: Tensor,
        token_ids: Tensor,
    ) -> float:
        """ Computes accuracy for comparing P(w|h1) > P(w|h2). """
        if self.use_full_model_probs:
            full_probs = self._decode(activations)
            counter_probs = self._decode(counter_activations)

            batch_size = full_probs.shape[0]
            probs = full_probs[range(batch_size), token_ids]
            counter_probs = counter_probs[range(batch_size), token_ids]
        else:
            probs = self._decode(activations, token_ids)
            counter_probs = self._decode(counter_activations, token_ids)

        return torch.mean((probs >= counter_probs).to(torch.float)).item()

    def _decode(
        self, activations: Tensor, token_ids: Optional[Tensor] = None
    ) -> Tensor:
        if hasattr(self.model, "decoder_w"):
            decoder_w = self.model.decoder_w
            decoder_b = self.model.decoder_b
            if token_ids is None:
                logits = activations @ decoder_w.t() + decoder_b

                return torch.nn.functional.log_softmax(logits, dim=-1)
            else:
                decoder_w = decoder_w[token_ids].unsqueeze(1)

                logits = torch.bmm(decoder_w, activations.unsqueeze(2)).squeeze()
                logits += decoder_b[token_ids]

                return logits
        elif hasattr(self.model, "decoder"):
            logits = getattr(self.model, "decoder")(activations)

            if token_ids is not None:
                batch_size = logits.size(0)
                logits = logits[range(batch_size), token_ids]

            return logits
        else:
            raise AttributeError("Model decoder not found")



import os
from itertools import product
from typing import Dict, Optional, Tuple, Union

import torch
from torch import Tensor
from torch.nn.functional import log_softmax
from torch.nn.utils.rnn import PackedSequence, pack_padded_sequence

from diagnnose.attribute import ShapleyTensor
from diagnnose.models.recurrent_lm import RecurrentLM
from diagnnose.typedefs.activations import (
    ActivationDict,
    ActivationNames,
    LayeredTensors,
)


class ForwardLSTM(RecurrentLM):
    """Defines a default uni-directional n-layer LSTM.

    Allows for extraction of intermediate states and gate activations.

    Parameters
    ----------
    state_dict : str
        Path to torch pickle containing the model parameter state dict.
    device : str, optional
        Torch device on which forward passes will be run.
        Defaults to cpu.
    rnn_name : str, optional
        Name of the rnn in the model state_dict. Defaults to `rnn`.
    encoder_name : str, optional
        Name of the embedding encoder in the model state_dict.
        Defaults to `encoder`.
    decoder_name : str, optional
        Name of the linear decoder in the model state_dict.
        Defaults to `decoder`.
    """

    ih_concat_order = ["h", "i"]
    split_order = ["i", "f", "g", "o"]

    def __init__(
        self,
        state_dict: str,
        device: str = "cpu",
        rnn_name: str = "rnn",
        encoder_name: str = "encoder",
        decoder_name: str = "decoder",
    ) -> None:
        super().__init__()
        print("Loading pretrained model...")

        with open(os.path.expanduser(state_dict), "rb") as mf:
            params: Dict[str, Tensor] = torch.load(mf, map_location=device)

        self.weight: LayeredTensors = {}
        self.bias: LayeredTensors = {}

        self._set_lstm_weights(params, rnn_name)

        # Encoder and decoder weights
        self.word_embeddings: Tensor = params[f"{encoder_name}.weight"]
        self.decoder_w: Tensor = params[f"{decoder_name}.weight"]
        self.decoder_b: Tensor = params[f"{decoder_name}.bias"]

        self.sizes[self.top_layer, "out"] = self.decoder_b.size(0)

        print("Model initialisation finished.")

    def forward(
        self,
        input_ids: Optional[Tensor] = None,
        inputs_embeds: Optional[Union[Tensor, ShapleyTensor]] = None,
        input_lengths: Optional[Tensor] = None,
        calc_causal_lm_probs: bool = False,
        compute_out: bool = False,
        only_return_top_embs: bool = False,
    ) -> Union[ActivationDict, Tensor]:
        if input_ids is not None and inputs_embeds is not None:
            raise ValueError(
                "You cannot specify both input_ids and inputs_embeds at the same time"
            )
        if inputs_embeds is None and input_ids is None:
            raise ValueError("inputs_embeds or input_ids must be provided")
        if inputs_embeds is None:
            inputs_embeds = self.create_inputs_embeds(input_ids)
        if len(inputs_embeds.shape) == 2:
            inputs_embeds = inputs_embeds.unsqueeze(0)

        inputs_embeds = inputs_embeds.to(self.device)

        iterator, unsorted_indices = self._create_iterator(inputs_embeds, input_lengths)

        all_activations = self._init_activations(inputs_embeds, compute_out)
        cur_activations = self.init_hidden(inputs_embeds.size(0))

        for w_idx, input_ in enumerate(iterator):
            num_input = input_.size(0)
            for a_name in cur_activations:
                cur_activations[a_name] = cur_activations[a_name][:num_input]

            cur_activations = self.forward_step(
                input_, cur_activations, compute_out=compute_out
            )

            for a_name in all_activations:
                all_activations[a_name][:num_input, w_idx] = cur_activations[a_name]

        # Batch had been sorted and needs to be unsorted to retain the original order
        for a_name, activations in all_activations.items():
            all_activations[a_name] = activations[unsorted_indices]

        if calc_causal_lm_probs:
            output_ids = input_ids[:, 1:].unsqueeze(-1)
            logits = all_activations[self.top_layer, "out"]
            probs = log_softmax(logits[:, :-1], dim=-1)
            all_activations[self.top_layer, "out"] = torch.gather(probs, -1, output_ids)

        if only_return_top_embs and compute_out:
            return all_activations[self.top_layer, "out"]
        elif only_return_top_embs:
            return all_activations[self.top_layer, "hx"]

        return all_activations

    def create_inputs_embeds(self, input_ids: Tensor) -> Tensor:
        return self.word_embeddings[input_ids]

    @staticmethod
    def _create_iterator(
        inputs_embeds: Tensor, input_lengths: Optional[Tensor]
    ) -> Tuple[Tuple[Tensor, ...], Tensor]:
        """Creates a PackedSequence that handles batching for the RNN.

        Batch items are sorted based on sentence length, allowing
        <pad> tokens to be skipped efficiently during the forward pass.

        Returns
        -------
        iterator : Tuple[Tensor, ...]
            Tuple of input tensors for each step in the sequence.
        unsorted_indices : Tensor
            Original order of the corpus prior to sorting.
        """
        if input_lengths is None:
            batch_size = inputs_embeds.size(0)
            input_lengths = torch.tensor(batch_size * [inputs_embeds.size(1)])

        packed_batch: PackedSequence = pack_padded_sequence(
            inputs_embeds, lengths=input_lengths, batch_first=True, enforce_sorted=False
        )

        iterator = torch.split(packed_batch.data, list(packed_batch.batch_sizes))

        return iterator, packed_batch.unsorted_indices

    def _init_activations(
        self, inputs_embeds: Tensor, compute_out: bool
    ) -> ActivationDict:
        """Returns a dictionary mapping activation names to tensors.

        If the input is a ShapleyTensor this dict will store the
        ShapleyTensors as well.

        Returns
        -------
        all_activations : ActivationDict
            Dictionary mapping activation names to tensors of shape:
            batch_size x max_sen_len x nhid.
        """
        batch_size, max_sen_len = inputs_embeds.shape[:2]
        all_activations: ActivationDict = {
            a_name: torch.zeros(batch_size, max_sen_len, self.nhid(a_name))
            for a_name in self.activation_names(compute_out)
        }

        if isinstance(inputs_embeds, ShapleyTensor):
            for a_name, activations in all_activations.items():
                all_activations[a_name] = ShapleyTensor(activations)

        return all_activations

    def forward_step(
        self,
        token_embeds: Tensor,
        prev_activations: ActivationDict,
        compute_out: bool = False,
    ) -> ActivationDict:
        """Performs a forward pass of one step across all layers.

        Parameters
        ----------
        token_embeds : Tensor
            Tensor of word embeddings at the current sentence position.
        prev_activations : ActivationDict
            Dict mapping the activation names of the previous hidden
            and cell states to their corresponding Tensors.
        compute_out : bool, optional
            Toggles the computation of the final decoder projection.
            If set to False this projection is not calculated.
            Defaults to True.

        Returns
        -------
        all_activations : ActivationDict
            Dictionary mapping activation names to tensors of shape:
            batch_size x max_sen_len x nhid.
        """
        cur_activations: ActivationDict = {}
        input_ = token_embeds

        for layer in range(self.num_layers):
            prev_hx = prev_activations[layer, "hx"]
            prev_cx = prev_activations[layer, "cx"]

            layer_activations = self.forward_cell(layer, input_, prev_hx, prev_cx)
            cur_activations.update(layer_activations)

            input_ = cur_activations[layer, "hx"]

        if compute_out:
            out = input_ @ self.decoder_w.t()
            out += self.decoder_b
            cur_activations[self.top_layer, "out"] = out

        return cur_activations

    def forward_cell(
        self, layer: int, input_: Tensor, prev_hx: Tensor, prev_cx: Tensor
    ) -> ActivationDict:
        """Performs the forward step of 1 LSTM cell.

        Parameters
        ----------
        layer : int
            Current RNN layer.
        input_ : Tensor
            Current input embedding. In higher layers this is h^l-1_t.
            Size: batch_size x nhid
        prev_hx : Tensor
            Previous hidden state. Size: batch_size x nhid
        prev_cx : Tensor
            Previous cell state. Size: batch_size x nhid

        Returns
        -------
        all_activations : ActivationDict
            Dictionary mapping activation names to tensors of shape:
            batch_size x max_sen_len x nhid.
        """
        # Shape: (bsz, nhid_h+emb_size)
        if self.ih_concat_order == ["h", "i"]:
            ih_concat = torch.cat((prev_hx, input_), dim=1)
        else:
            ih_concat = torch.cat((input_, prev_hx), dim=1)

        # Shape: (bsz, 4*nhid_c)
        proj = ih_concat @ self.weight[layer]
        if layer in self.bias:
            proj += self.bias[layer]

        split_proj: Dict[str, Tensor] = dict(
            zip(self.split_order, torch.split(proj, self.sizes[layer, "cx"], dim=1))
        )

        # Shapes: (bsz, nhid_c)
        f_g = torch.sigmoid(split_proj["f"])
        i_g = torch.sigmoid(split_proj["i"])
        o_g = torch.sigmoid(split_proj["o"])
        c_tilde_g = torch.tanh(split_proj["g"])

        cx = f_g * prev_cx + i_g * c_tilde_g
        hx = o_g * torch.tanh(cx)

        activation_dict = {
            (layer, "hx"): hx,
            (layer, "cx"): cx,
            (layer, "f_g"): f_g,
            (layer, "i_g"): i_g,
            (layer, "o_g"): o_g,
            (layer, "c_tilde_g"): c_tilde_g,
        }

        if layer == 0:
            activation_dict[0, "emb"] = input_

        return activation_dict

    def _set_lstm_weights(self, params: Dict[str, Tensor], rnn_name: str) -> None:
        # LSTM weights
        layer = 0
        # 1 layer RNNs do not have a layer suffix in the state_dict
        no_suffix = self.param_names(0, rnn_name, no_suffix=True)["weight_hh"] in params
        assert (
            self.param_names(0, rnn_name, no_suffix=no_suffix)["weight_hh"] in params
        ), "rnn weight name not found, check if setup is correct"

        while (
            self.param_names(layer, rnn_name, no_suffix=no_suffix)["weight_hh"]
            in params
        ):
            param_names = self.param_names(layer, rnn_name, no_suffix=no_suffix)

            w_h = params[param_names["weight_hh"]]
            w_i = params[param_names["weight_ih"]]

            # Shape: (emb_size+nhid_h, 4*nhid_c)
            self.weight[layer] = torch.cat((w_h, w_i), dim=1).t()

            if param_names["bias_hh"] in params:
                # Shape: (4*nhid_c,)
                self.bias[layer] = (
                    params[param_names["bias_hh"]] + params[param_names["bias_ih"]]
                )

            self.sizes.update(
                {
                    (layer, "emb"): w_i.size(1),
                    (layer, "hx"): w_h.size(1),
                    (layer, "cx"): w_h.size(1),
                }
            )
            layer += 1

            if no_suffix:
                break

    @staticmethod
    def param_names(
        layer: int, rnn_name: str, no_suffix: bool = False
    ) -> Dict[str, str]:
        """Creates a dictionary of parameter names in a state_dict.

        Parameters
        ----------
        layer : int
            Current layer index.
        rnn_name : str
            Name of the rnn in the model state_dict. Defaults to `rnn`.
        no_suffix : bool, optional
            Toggle to omit the `_l{layer}` suffix from a parameter name.
            1-layer RNNs do not have this suffix. Defaults to False.

        Returns
        -------
        param_names : Dict[str, str]
            Dictionary mapping a general parameter name to the model
            specific parameter name.
        """
        if no_suffix:
            return {
                "weight_hh": f"{rnn_name}.weight_hh",
                "weight_ih": f"{rnn_name}.weight_ih",
                "bias_hh": f"{rnn_name}.bias_hh",
                "bias_ih": f"{rnn_name}.bias_ih",
            }

        return {
            "weight_hh": f"{rnn_name}.weight_hh_l{layer}",
            "weight_ih": f"{rnn_name}.weight_ih_l{layer}",
            "bias_hh": f"{rnn_name}.bias_hh_l{layer}",
            "bias_ih": f"{rnn_name}.bias_ih_l{layer}",
        }

    def activation_names(self, compute_out: bool = False) -> ActivationNames:
        """Returns a list of all the model's activation names.

        Parameters
        ----------
        compute_out : bool, optional
            Toggles the computation of the final decoder projection.
            If set to False this projection is not calculated.
            Defaults to True.

        Returns
        -------
        activation_names : ActivationNames
            List of (layer, name) tuples.
        """
        lstm_names = ["hx", "cx", "f_g", "i_g", "o_g", "c_tilde_g"]

        activation_names = list(product(range(self.num_layers), lstm_names))
        activation_names.append((0, "emb"))

        if compute_out:
            activation_names.append((self.top_layer, "out"))

        return activation_names



import os
from typing import Dict, List, Optional

from diagnnose.corpus import Corpus
from diagnnose.typedefs.syntax import SyntaxEvalCorpora

from ..task import SyntaxEvalTask


class LakretzTask(SyntaxEvalTask):
    descriptions = {
        "adv": {"items_per_condition": 900, "conditions": ["S", "P"]},
        "adv_adv": {"items_per_condition": 900, "conditions": ["S", "P"]},
        "adv_conjunction": {"items_per_condition": 600, "conditions": ["S", "P"]},
        "namepp": {"items_per_condition": 900, "conditions": ["S", "P"]},
        "nounpp": {"items_per_condition": 600, "conditions": ["SS", "SP", "PS", "PP"]},
        "nounpp_adv": {
            "items_per_condition": 600,
            "conditions": ["SS", "SP", "PS", "PP"],
        },
        "simple": {"items_per_condition": 300, "conditions": ["S", "P"]},
    }

    def initialize(
        self, path: str, subtasks: Optional[List[str]] = None
    ) -> SyntaxEvalCorpora:
        """Performs the initialization for the tasks of
        Marvin & Linzen (2018)

        Arxiv link: https://arxiv.org/pdf/1808.09031.pdf

        Repo: https://github.com/BeckyMarvin/LM_syneval

        Parameters
        ----------
        path : str
            Path to directory containing the Marvin datasets that can be
            found in the github repo.
        subtasks : List[str], optional
            The downstream tasks that will be tested. If not provided this
            will default to the full set of conditions.

        Returns
        -------
        corpora : Dict[str, Corpus]
            Dictionary mapping a subtask to a Corpus.
        """
        subtasks = subtasks or self.descriptions.keys()

        corpora: SyntaxEvalCorpora = {}

        for subtask in subtasks:
            items_per_condition = self.descriptions[subtask]["items_per_condition"]

            for i, condition in enumerate(self.descriptions[subtask]["conditions"]):
                start_idx = i * items_per_condition
                stop_idx = (i + 1) * items_per_condition
                condition_slice = slice(start_idx, stop_idx)

                corpus = self._create_corpus(
                    os.path.join(path, f"{subtask}.txt"), condition_slice
                )

                corpora.setdefault(subtask, {})[condition] = corpus

        return corpora

    def _create_corpus(self, path: str, condition_slice: slice) -> Corpus:
        """Attach the correct and incorrect verb form to each sentence
        in the corpus.
        """
        raw_corpus = Corpus.create_raw_corpus(path)

        for idx in range(0, len(raw_corpus), 2):
            token = raw_corpus[idx][0].split()[-1]
            counter_token = raw_corpus[idx + 1][0].split()[-1]
            sen = " ".join(raw_corpus[idx][0].split()[:-1])
            raw_corpus[idx] = [sen, token, counter_token]

        raw_corpus = raw_corpus[::2][condition_slice]

        fields = Corpus.create_fields(
            ["sen", "token", "counter_token"], tokenizer=self.tokenizer
        )

        examples = Corpus.create_examples(raw_corpus, fields)

        return Corpus(examples, fields)



import os
from typing import Dict, List, Optional

from torchtext.data import RawField

from diagnnose.corpus import Corpus
from diagnnose.typedefs.syntax import SyntaxEvalCorpora

from ..task import SyntaxEvalTask


class WinobiasTask(SyntaxEvalTask):
    def initialize(
        self, path: str, subtasks: Optional[List[str]] = None
    ) -> SyntaxEvalCorpora:
        """

        Parameters
        ----------
        path : str
            Path to directory containing the Marvin datasets that can be
            found in the github repo.
        subtasks : List[str], optional
            The downstream tasks that will be tested. If not provided this
            will default to the full set of conditions.

        Returns
        -------
        corpora : Dict[str, Corpus]
            Dictionary mapping a subtask to a Corpus.
        """
        subtasks = subtasks or ["stereo", "unamb"]

        corpora: SyntaxEvalCorpora = {}

        for subtask in subtasks:
            for condition in ["FF", "FM", "MF", "MM"]:
                corpus = Corpus.create(
                    os.path.join(path, f"{subtask}_{condition}.tsv"),
                    header_from_first_line=True,
                    tokenizer=self.tokenizer,
                )

                self._add_output_classes(corpus)

                corpora.setdefault(subtask, {})[condition] = corpus

        return corpora

    @staticmethod
    def _add_output_classes(corpus: Corpus) -> None:
        """ Set the the pronouns for each sentence. """
        corpus.fields["token"] = RawField()
        corpus.fields["counter_token"] = RawField()

        corpus.fields["token"].is_target = False
        corpus.fields["counter_token"].is_target = False

        for ex in corpus:
            setattr(ex, "token", "he")
            setattr(ex, "counter_token", "she")



import os
import pickle
from typing import Iterator, Optional, Union

import torch
from torch import Tensor

from diagnnose.activations.activation_index import activation_index_to_iterable
from diagnnose.typedefs.activations import (
    ActivationDict,
    ActivationKey,
    ActivationName,
    ActivationNames,
    ActivationRanges,
    SelectionFunc,
)
from diagnnose.utils.pickle import load_pickle


class ActivationReader:
    """Reads in pickled activations that have been extracted.

    An ``ActivationReader`` can also be created directly from an
    ``ActivationDict``, in which case the corresponding
    ``ActivationRanges`` and ``SelectionFunc`` should be provided too.

    Parameters
    ----------
    activations_dir : str, optional
        Directory containing the extracted activations.
    activation_dict : ActivationDict, optional
        If activations have not been extracted to disk, the
        activation_dict containing all extracted activations can be
        provided directly as well.
    activation_names : ActivationNames, optional
        Activation names, provided as a list of ``(layer, name)``
        tuples. If not provided the index to
        :func:`~diagnnose.activations.ActivationReader.__getitem__`
        must always contain the activation_name that is being requested,
        as the ``ActivationReader`` can not infer it automatically.
    activation_ranges : ActivationRanges, optional
        ``ActivationRanges`` dictionary that should be provided if
        ``activation_dict`` is passed directly.
    selection_func : SelectionFunc, optional
        ``SelectionFunc`` that was used for extraction and that should
        be passed if ``activation_dict`` is passed directly.
    store_multiple_activations : bool, optional
        Set to true to store multiple activation arrays in RAM at once.
        Defaults to False, meaning that only one activation type will be
        stored in the class.
    cat_activations : bool, optional
        Toggle to concatenate the activations returned by
        :func:`~diagnnose.activations.ActivationReader.__getitem__`.
        Otherwise the activations will be split into a tuple with each
        each tuple item containing the activations of one sentence.
    """

    def __init__(
        self,
        activations_dir: Optional[str] = None,
        activation_dict: Optional[ActivationDict] = None,
        activation_names: Optional[ActivationNames] = None,
        activation_ranges: Optional[ActivationRanges] = None,
        selection_func: Optional[SelectionFunc] = None,
        store_multiple_activations: bool = False,
        cat_activations: bool = False,
    ) -> None:
        if activations_dir is not None:
            assert os.path.exists(
                activations_dir
            ), f"Activations dir not found: {activations_dir}"
            assert (
                activation_dict is None
            ), "activations_dir and activations_dict can not be provided simultaneously"
        else:
            assert activation_dict is not None
            assert activation_ranges is not None
            assert selection_func is not None

        self.activations_dir = activations_dir
        self.activation_dict: ActivationDict = activation_dict or {}
        self.activation_names: ActivationNames = activation_names or list(
            self.activation_dict.keys()
        )

        self._activation_ranges: Optional[ActivationRanges] = activation_ranges
        self._selection_func: Optional[SelectionFunc] = selection_func

        self.store_multiple_activations = store_multiple_activations
        self.cat_activations = cat_activations

    def __getitem__(self, key: ActivationKey) -> Union[Tensor, Iterator[Tensor]]:
        """Allows for concise and efficient indexing of activations.

        The ``key`` argument should be either an ``ActivationIndex``
        (i.e. an iterable that can be used to index a tensor), or a
        ``(index, activation_name)`` tuple. An ``activation_name`` is
        a tuple of shape ``(layer, name)``.

        If multiple activation_names have been extracted the
        ``activation_name`` must be provided, otherwise it can be left
        out.

        The return value is a generator of tensors, with each tensor of
        shape (sen_len, nhid), or a concatenated tensor if
        ``self.cat_activations`` is set to ``True``.

        Example usage:

        .. code-block:: python

            activation_reader = ActivationReader(
                dir, activation_names=[(0, "hx"), (1, "hx")], **kwargs
            )

            # activation_name must be passed because ActivationReader
            # contains two activation_names.
            activations_first_sen = activation_reader[0, (1, "hx")]
            all_activations = activation_reader[:, (1, "hx")]


            activation_reader2 = ActivationReader(
                dir, activation_names=[(1, "hx")], **kwargs
            )

            # activation_name can be left implicit.
            activations_first_10_sens = activation_reader2[:10]

        Parameters
        ----------
        key : ActivationKey
            ``ActivationIndex`` or ``(index, activation_name)``, as
            explained above.

        Returns
        -------
        split_activations : Tensor | Iterator[Tensor, ...]
            Tensor, if ``self.cat_activations`` is set to True.
            Otherwise a Generator of tensors, with each item
            corresponding to the extracted activations of a specific
            sentence.

        .. automethod:: __getitem__
        """
        if isinstance(key, tuple):
            index, activation_name = key
        else:
            assert (
                len(self.activation_names) == 1
            ), "Activation name must be provided if multiple activations have been extracted"
            index = key
            activation_name = self.activation_names[0]

        iterable_index = activation_index_to_iterable(
            index, len(self.activation_ranges)
        )
        ranges = [self.activation_ranges[idx] for idx in iterable_index]

        sen_indices = torch.cat([torch.arange(*r) for r in ranges]).to(torch.long)

        if activation_name not in self.activation_dict:
            self._read_activations(activation_name)

        if self.cat_activations:
            return self.activation_dict[activation_name][sen_indices]

        return self.get_item_generator(ranges, self.activation_dict[activation_name])

    @staticmethod
    def get_item_generator(ranges, activations) -> Iterator[Tensor]:
        for start, stop in ranges:
            yield activations[start:stop]

    def __len__(self) -> int:
        """ Returns the total number of extracted activations. """
        return self.activation_ranges[-1][1]

    def to(self, device: str) -> None:
        """ Cast activations to a different device. """
        self.activation_dict = {
            a_name: activation.to(device)
            for a_name, activation in self.activation_dict.items()
        }

    @property
    def activation_ranges(self) -> ActivationRanges:
        if self._activation_ranges is None:
            ranges_path = os.path.join(self.activations_dir, "activation_ranges.pickle")
            self._activation_ranges = load_pickle(ranges_path)
        return self._activation_ranges

    @property
    def selection_func(self) -> SelectionFunc:
        if self._selection_func is None:
            selection_func_path = os.path.join(
                self.activations_dir, "selection_func.dill"
            )
            self._selection_func = load_pickle(selection_func_path, use_dill=True)
        return self._selection_func

    def activations(self, activation_name: ActivationName) -> Tensor:
        activations = self.activation_dict.get(activation_name, None)

        if activations is None:
            activations = self._read_activations(activation_name)
            if not self.store_multiple_activations:
                self.activation_dict = {}  # reset activation_dict
            self.activation_dict[activation_name] = activations

        return activations

    def _read_activations(self, activation_name: ActivationName) -> None:
        """Reads the pickled activations of activation_name

        Parameters
        ----------
        activation_name : ActivationName
            (layer, name) tuple indicating the activations to be read in

        Returns
        -------
        activations : Tensor
            Torch tensor of activation values
        """
        layer, name = activation_name
        filename = os.path.join(self.activations_dir, f"{layer}-{name}.pickle")

        activations = None

        n = 0

        # The activations are stored as a series of pickle dumps, and
        # are therefore loaded until an EOFError is raised.
        with open(filename, "rb") as f:
            while True:
                try:
                    sen_activations = pickle.load(f)

                    # To make hidden size dependent of data only, the activations array
                    # is created only after observing the first batch of activations.
                    if activations is None:
                        hidden_size = sen_activations.shape[1]
                        activations = torch.empty(
                            (len(self), hidden_size),
                            dtype=sen_activations.dtype,
                            device=sen_activations.device,
                        )

                    i = len(sen_activations)
                    activations[n : n + i] = sen_activations
                    n += i
                except EOFError:
                    break

        assert activations is not None, (
            f"Reading activations [{layer}, {name}] returned None, "
            f"check if file exists and is non-empty."
        )

        if not self.store_multiple_activations:
            self.activation_dict = {}  # reset activation_dict
        self.activation_dict[activation_name] = activations



from typing import List, Tuple, Union

from torch import Tensor
from transformers import BatchEncoding, PreTrainedTokenizer

from .decomposer import Decomposer


class Explainer:
    """ Generates an explanation for a specific input. """

    def __init__(
        self,
        decomposer: Decomposer,
        tokenizer: PreTrainedTokenizer,
    ):
        self.decomposer = decomposer
        self.tokenizer = tokenizer

    def explain(self, input_tokens: Union[str, List[str]], output_tokens: List[str]):
        batch_encoding = self._tokenize(input_tokens)

        out, contributions = self.decomposer.decompose(batch_encoding)

        output_ids, mask_ids = self._create_output_ids(batch_encoding, output_tokens)

        full_probs = self._fetch_token_probs(out, output_ids, mask_ids)
        contribution_probs = [
            self._fetch_token_probs(contribution, output_ids, mask_ids)
            for contribution in contributions
        ]

        return full_probs, contribution_probs

    def _tokenize(self, input_tokens: Union[str, List[str]]) -> BatchEncoding:
        input_tokens = [input_tokens] if isinstance(input_tokens, str) else input_tokens

        batch_encoding = self.tokenizer(
            input_tokens,
            padding=True,
            return_attention_mask=False,
            return_length=True,
            return_token_type_ids=False,
        )

        return batch_encoding

    def _create_output_ids(
        self, batch_encoding: BatchEncoding, output_tokens: List[str]
    ) -> Tuple[List[int], List[int]]:
        mask_idx = self.tokenizer.convert_tokens_to_ids(self.tokenizer.mask_token)
        unk_idx = self.tokenizer.convert_tokens_to_ids(self.tokenizer.unk_token)
        mask_idx = None if mask_idx == unk_idx else mask_idx
        mask_ids = [
            sen.index(mask_idx) if mask_idx in sen else final_idx - 1
            for sen, final_idx in zip(
                batch_encoding["input_ids"], batch_encoding["length"]
            )
        ]

        output_ids = []
        mask_token = self.tokenizer.mask_token or "<mask>"
        for token in output_tokens:
            sub_token_id = self.tokenizer.convert_tokens_to_ids(token)

            if sub_token_id == unk_idx:
                # Simply encoding "token" often yields a different index than when it is embedded
                # within a sentence, hence the ugly hack here.
                sub_token_ids = self.tokenizer.encode(
                    f"{mask_token} {token}", add_special_tokens=False
                )[1:]
                sub_tokens = self.tokenizer.convert_ids_to_tokens(sub_token_ids)
                assert (
                    len(sub_token_ids) == 1
                ), f"Multi-subword tokens not supported ({token} -> {str(sub_tokens)})"
                sub_token_id = sub_token_ids[0]

            output_ids.append(sub_token_id)

        return output_ids, mask_ids

    @staticmethod
    def _fetch_token_probs(
        probs: Tensor, output_ids: List[int], mask_ids: List[int]
    ) -> Tensor:
        """Fetches the probability of each output class at the position
        of the corresponding mask_idx.

        Parameters
        ----------
        probs : Tensor
            Tensor with output probabilities of shape:
            batch_size x max_sen_len x output_dim.
        output_ids : List[int]
            List of indices of the output classes that are decomposed.
        mask_ids : List[int]
            List of indices that signify the position of each sentence
            in the input batch where the decomposition will take place.

        Returns
        -------
        token_probs : Tensor
            Tensor containing the corresponding probabilities.
        """
        mask_probs = probs[range(probs.size(0)), mask_ids]
        token_probs = mask_probs[:, output_ids]

        return token_probs

    def print_attributions(
        self,
        full_probs: Tensor,
        contribution_probs: List[Tensor],
        input_tokens: Union[str, List[str]],
        output_tokens: List[str],
    ):
        batch_encoding = self._tokenize(input_tokens)

        for sen_idx, token_ids in enumerate(batch_encoding["input_ids"]):
            print((" " * 15) + "".join(f"{w:<15}" for w in output_tokens))
            print(
                f"{'Full logits':<15}"
                + "".join(f"{p:<15.3f}" for p in full_probs[sen_idx])
            )
            print("-" * 15 * (len(output_tokens) + 1))
            sen_features = [self.tokenizer.decode([w]) for w in token_ids]
            sen_len = batch_encoding["length"][sen_idx]
            features = ["model_bias", *sen_features[:sen_len]]
            for i, feature in enumerate(features):
                print(
                    f"{feature:<15}"
                    + "".join(f"{p:<15.3f}" for p in contribution_probs[i][sen_idx])
                )
            print("\n")



from typing import Callable, Optional

from torchtext.data import Iterator

from diagnnose.corpus import Corpus


def create_iterator(
    corpus: Corpus, batch_size: int = 1, device: str = "cpu", sort: bool = False
) -> Iterator:
    """Transforms a Corpus into an :class:`torchtext.data.Iterator`.

    Parameters
    ----------
    corpus : Corpus
         Corpus containing sentences that will be tokenized and
         transformed into a batch.
    batch_size : int, optional
        Amount of sentences processed per forward step. Higher batch
        size increases processing speed, but should be done
        accordingly to the amount of available RAM. Defaults to 1.
    device : str, optional
        Torch device on which forward passes will be run.
        Defaults to cpu.
    sort : bool, optional
        Toggle to sort the corpus based on sentence length. Defaults to
        ``False``.

    Returns
    -------
    iterator : Iterator
        Iterator containing the batchified Corpus.
    """
    if sort is not None:
        sort_key: Optional[Callable] = lambda ex: len(getattr(ex, corpus.sen_column))
    else:
        sort_key = None

    iterator = Iterator(
        dataset=corpus,
        batch_size=batch_size,
        device=device,
        shuffle=False,
        sort=sort,
        sort_key=sort_key,
        train=False,
    )

    return iterator



import torch

from .shapley_tensor import ShapleyTensor


class GCDTensor(ShapleyTensor):
    def mul_contributions(self, *args, **kwargs):
        gate, source = args

        if not isinstance(gate, ShapleyTensor):
            contributions = [
                torch.mul(gate, contribution, **kwargs)
                for contribution in source.contributions
            ]
        elif not isinstance(source, ShapleyTensor):
            contributions = [
                torch.mul(contribution, source, **kwargs)
                for contribution in gate.contributions
            ]
        else:
            gate_contributions = sum(gate.contributions)
            contributions = [
                torch.mul(gate_contributions, source_contribution, **kwargs)
                for source_contribution in source.contributions
            ]

        return contributions



import os
import pickle
import warnings
from contextlib import ExitStack
from typing import BinaryIO, Optional

import dill

from diagnnose.typedefs.activations import (
    ActivationDict,
    ActivationFiles,
    ActivationNames,
    ActivationRanges,
    SelectionFunc,
)
from diagnnose.utils.pickle import dump_pickle

from .activation_reader import ActivationReader


class ActivationWriter:
    """Writes activations to file, using an ExitStack.

    Parameters
    ----------
    activations_dir : str, optional
        Directory to which activations will be written
    """

    def __init__(self, activations_dir: str) -> None:
        self.activations_dir = activations_dir

        self.activation_names: ActivationNames = []
        self.activation_files: ActivationFiles = {}
        self.activation_ranges_file: Optional[BinaryIO] = None
        self.selection_func_file: Optional[BinaryIO] = None

    def create_output_files(
        self, stack: ExitStack, activation_names: ActivationNames
    ) -> None:
        """ Opens a file for each to-be-extracted activation. """
        self.activation_names = activation_names

        if not os.path.exists(self.activations_dir):
            os.makedirs(self.activations_dir)

        if os.listdir(self.activations_dir):
            warnings.warn("Output directory %s is not empty" % self.activations_dir)

        self.activation_files = {
            (layer, name): stack.enter_context(
                open(os.path.join(self.activations_dir, f"{layer}-{name}.pickle"), "wb")
            )
            for (layer, name) in self.activation_names
        }

        self.activation_ranges_file = stack.enter_context(
            open(os.path.join(self.activations_dir, "activation_ranges.pickle"), "wb")
        )
        self.selection_func_file = stack.enter_context(
            open(os.path.join(self.activations_dir, "selection_func.dill"), "wb")
        )

    def dump_activations(self, activations: ActivationDict) -> None:
        """Dumps the generated activations to a list of opened files

        Parameters
        ----------
        activations : PartialArrayDict
            The Tensors for each activation that was specifed by
            self.activation_names at initialization.
        """
        for activation_name in self.activation_names:
            assert (
                activation_name in self.activation_files.keys()
            ), "Activation file is not opened"

            pickle.dump(
                activations[activation_name], self.activation_files[activation_name]
            )

    def dump_meta_info(
        self, activation_ranges: ActivationRanges, selection_func: SelectionFunc
    ) -> None:
        """ Dumps activation_ranges and selection_func to disk. """
        assert self.activation_ranges_file is not None
        assert self.selection_func_file is not None

        pickle.dump(activation_ranges, self.activation_ranges_file)
        dill.dump(selection_func, self.selection_func_file, recurse=True)

    def concat_pickle_dumps(self, overwrite: bool = True) -> None:
        """Concatenates a sequential pickle dump and pickles to file .

        Note that this overwrites the sequential pickle dump by default.

        Parameters
        ----------
        overwrite : bool, optional
            Set to True to overwrite the file containing the sequential
            pickle dump, otherwise creates a new file. Defaults to True.
        """
        activation_reader = ActivationReader(
            self.activations_dir, store_multiple_activations=False
        )

        for (layer, name) in self.activation_names:
            activations = activation_reader.activations((layer, name))
            filename = os.path.join(self.activations_dir, f"{name}_l{layer}.pickle")
            if not overwrite:
                filename = filename.replace(".pickle", "_concat.pickle")
            dump_pickle(activations, filename)
            del activations



from functools import reduce
from typing import Callable, List, Optional, Union

import torch
from torch import Tensor
from torch.nn.functional import log_softmax
from transformers import (
    AutoModel,
    AutoModelForCausalLM,
    AutoModelForMaskedLM,
    AutoModelForQuestionAnswering,
    AutoModelForSequenceClassification,
    AutoModelForTokenClassification,
    DistilBertForMaskedLM,
    PreTrainedModel,
    XLMWithLMHeadModel,
)

from diagnnose.attribute import ShapleyTensor
from diagnnose.models import LanguageModel
from diagnnose.typedefs.activations import (
    ActivationDict,
    ActivationName,
    ActivationNames,
)

mode_to_auto_model = {
    "causal_lm": AutoModelForCausalLM,
    "masked_lm": AutoModelForMaskedLM,
    "question_answering": AutoModelForQuestionAnswering,
    "sequence_classification": AutoModelForSequenceClassification,
    "token_classification": AutoModelForTokenClassification,
}


class TransformerLM(LanguageModel):
    """Huggingface LM wrapper.

    Parameters
    ----------
    transformer_type: str
        Transformer type that can be passed to
        ``auto_model.from_pretrained`` as one of the valid models made
        available by Huggingface, e.g. ``"roberta-base"``.
    mode : str, optional
        Language model mode, one of ``"causal_lm"``, ``"masked_lm"``,
        ``"question_answering"``, ``"sequence_classification"``, or
        ``"token_classification"``. If not provided the model will be
        imported using ``AutoModel``, which often yields an LM with no
        task-specific head on top.
    embeddings_attr : str, optional
        Attribute name of the word embeddings of the model. Can be
        nested. For example, if the word embeddings are stored as an
        ``"wte"`` attribute that is part of the ``"encoder"`` attribute
        of the full model, you would pass ``"encoder.wte"``. For the
        following models this parameter does not need to be passed:
        ``"(distil)-(Ro)BERT(a)"``, ``"(distil)-gpt2"``, ``"XLM"``, ``""``, ``""``,
    device : str, optional
        Torch device on which forward passes will be run.
        Defaults to cpu.
    """

    def __init__(
        self,
        transformer_type: str,
        mode: Optional[str] = None,
        embeddings_attr: Optional[str] = None,
        cache_dir: Optional[str] = None,
        device: str = "cpu",
    ):
        super().__init__()

        auto_model = mode_to_auto_model.get(mode, AutoModel)

        self.pretrained_model: PreTrainedModel = auto_model.from_pretrained(
            transformer_type, cache_dir=cache_dir
        )

        self.embeddings_attr = embeddings_attr
        self.device = device

    def forward(
        self,
        input_ids: Optional[Union[Tensor, List[int]]] = None,
        inputs_embeds: Optional[Union[Tensor, ShapleyTensor]] = None,
        input_lengths: Optional[List[int]] = None,
        attention_mask: Optional[Union[Tensor, List[int]]] = None,
        compute_out: bool = True,
        calc_causal_lm_probs: bool = False,
        only_return_top_embs: bool = True,
    ) -> Union[ActivationDict, Tensor]:
        if input_ids is not None and inputs_embeds is not None:
            raise ValueError(
                "You cannot specify both input_ids and inputs_embeds at the same time"
            )
        if inputs_embeds is None and input_ids is None:
            raise ValueError("inputs_embeds or input_ids must be provided")
        if inputs_embeds is None:
            inputs_embeds = self.create_inputs_embeds(input_ids)
        if len(inputs_embeds.shape) == 2:
            inputs_embeds = inputs_embeds.unsqueeze(0)  # Add batch dimension
        if input_lengths is None:
            batch_size, max_sen_len = inputs_embeds.shape[:2]
            input_lengths = torch.tensor(batch_size * [max_sen_len], device=self.device)
        if isinstance(attention_mask, list):
            attention_mask = torch.tensor(attention_mask, device=self.device)
        if attention_mask is None:
            attention_mask = self.create_attention_mask(input_lengths)

        model = (
            self.pretrained_model if compute_out else self.pretrained_model.base_model
        )
        inputs_embeds = inputs_embeds.to(self.device)
        attention_mask = attention_mask.to(self.device)

        output = model(inputs_embeds=inputs_embeds, attention_mask=attention_mask)

        if hasattr(output, "logits"):
            logits: Tensor = output.logits
            activation_name = (-1, "out")
        elif hasattr(output, "last_hidden_state"):
            logits = output.last_hidden_state
            activation_name = (-1, "hx")
        else:
            raise AttributeError

        if calc_causal_lm_probs:
            output_ids = input_ids[:, 1:].unsqueeze(-1)
            probs = log_softmax(logits[:, :-1], dim=-1)
            logits = torch.gather(probs, -1, output_ids)

        if only_return_top_embs:
            return logits

        return {activation_name: logits}

    def create_attention_mask(self, input_lengths: List[int]) -> Tensor:
        """Creates an attention mask as described in:
        https://huggingface.co/transformers/glossary.html#attention-mask

        Parameters
        ----------
        input_lengths : List[int]
            List containing sentence lengths of each batch item.

        Returns
        -------
        attention_mask : Tensor
            Attention mask prescribing which items may be taken into
            account by the attention mechanism.
            Size: batch_size x max_sen_length
        """
        max_sen_len = max(input_lengths)

        attention_mask = torch.zeros(
            len(input_lengths), max_sen_len, device=self.device
        )

        for idx, length in enumerate(input_lengths):
            attention_mask[idx, :length] = 1.0

        return attention_mask

    @property
    def embeddings(self) -> Callable[[Tensor], Tensor]:
        if self.embeddings_attr is not None:
            attrs = self.embeddings_attr.split(".")
            return reduce(getattr, attrs, self.pretrained_model)
        else:
            base_model = self.pretrained_model.base_model
            if hasattr(base_model, "wte"):
                # GPT-2
                return base_model.wte
            elif hasattr(base_model, "embeddings"):
                if hasattr(base_model.embeddings, "word_embeddings"):
                    # BERT-based models, Electra, Longformer, Reformer
                    return base_model.embeddings.word_embeddings
                else:
                    # XLM
                    return base_model.embeddings
            elif hasattr(base_model, "word_embedding"):
                # XLNet
                return base_model.word_embedding
            elif hasattr(base_model, "w"):
                # CTRL
                return base_model.w
            elif hasattr(base_model, "encoder"):
                # T5
                return base_model.encoder.embed_tokens
            else:
                raise AttributeError("word embedding attribute not found")

    def create_inputs_embeds(self, input_ids: Union[Tensor, List[int]]) -> Tensor:
        if isinstance(input_ids, list):
            input_ids = torch.tensor(input_ids, device=self.device)

        inputs_embeds = self.embeddings(input_ids)

        return inputs_embeds

    @property
    def decoder(self) -> torch.nn.Module:
        # RoBERTa / BERT
        for attr in ["lm_head", "cls"]:
            if hasattr(self.pretrained_model, attr):
                return getattr(self.pretrained_model, attr)

        if isinstance(self.pretrained_model, DistilBertForMaskedLM):
            return torch.nn.Sequential(
                self.pretrained_model.vocab_transform,
                torch.nn.GELU(),
                self.pretrained_model.vocab_layer_norm,
                self.pretrained_model.vocab_projector,
            )

        if isinstance(self.pretrained_model, XLMWithLMHeadModel):
            return self.pretrained_model.pred_layer.proj

        raise AttributeError("Model decoder not found")

    @property
    def num_layers(self) -> int:
        return self.pretrained_model.config.n_layer

    @property
    def top_layer(self) -> int:
        return -1

    def nhid(self, activation_name: ActivationName) -> int:
        if activation_name[1] == "out":
            return self.pretrained_model.config.vocab_size

        return self.pretrained_model.config.hidden_size

    def activation_names(self) -> ActivationNames:
        return [(-1, "out")]



from typing import Dict, Union

from diagnnose.corpus import Corpus

# subtask -> Corpus | (condition -> Corpus)
SyntaxEvalCorpora = Dict[str, Union[Corpus, Dict[str, Corpus]]]
# subtask -> accuracy | (condition -> accuracy)
ResultsDict = Dict[str, Union[float, Dict[str, float]]]



import re
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple, Union

# sen_id -> (licensor, scope, npi_present) -> item
ItemCondition = Tuple[int, int, int]
CorpusItem = Dict[ItemCondition, Dict[str, Any]]
CorpusDict = Dict[int, CorpusItem]

ENVS = [
    "adverbs",
    "conditional",
    "determiner_negation_biclausal",
    "only",
    "quantifier",
    "questions",
    "sentential_negation_biclausal",
    "simplequestions",
    "superlative",
]


def preproc_warstadt(path: str) -> CorpusDict:
    """Reads and preprocesses the NPI corpus of Warstadt et al. (2019).

    Paper: https://arxiv.org/pdf/1901.03438.pdf

    Data: https://alexwarstadt.files.wordpress.com/2019/08/npi_lincensing_data.zip

    Parameters
    ----------
    path : str
        Path to .tsv corpus file.

    Returns
    -------
    sen_id2items : CorpusDict
        Dictionary mapping a sen_id to a triplet (licensor, scope,
        npi_present) to the full corpus item.
    env2sen_ids : EnvIdDict
        Dictionary mapping each env type to a list of sen_id's of that
        type.
    """
    with open(path) as f:
        lines = [line[:-1].split("\t") for line in f]

    def preproc(s):
        return int(s) if s.isnumeric() else s

    # Separate punctuation from adjacent tokens
    pattern = re.compile(r"[.,?]")

    def preproc_sen(s):
        return pattern.sub(lambda m: f" {m.group(0)}", s).split()

    # map each line to a dictionary
    raw_items: List[Dict[str, Any]] = [
        {
            **{k: preproc(v) for k, v in [x.split("=") for x in line[0].split("-")]},
            "correct": bool(int(line[1])),
            "sen": preproc_sen(line[-1]),
        }
        for line in lines
    ]

    extra_idx = 0

    for idx, item in enumerate(raw_items):
        # Remove apostrophes that wrap the full sentence.
        if item["sen"][0][0] == '"':
            item["sen"][0] = item["sen"][0][1:]
        if item["sen"][-1][-1] == '"':
            item["sen"][-1] = item["sen"][-1][:-1]

        sen_id = idx // 8
        # There exist only 4 instead of 8 conditions for `simplequestions`.
        if item["env"] == "simplequestions":
            extra_idx += idx % 8 == 4
        item["sen_id"] = sen_id + extra_idx

        # Cut multi-word NPIs into pieces.
        if item["npi_present"] == 1:
            if item["npi"] in ["atall", "inyears"]:
                item["npi"] = f"{item['npi'][:2]} {item['npi'][2:]}"

    sen_id2items = defaultdict(dict)

    for item in raw_items:
        sen_id2items[item["sen_id"]][
            item["licensor"], item["scope"], item["npi_present"]
        ] = item

    return sen_id2items


def create_downstream_corpus(
    orig_corpus: Union[str, CorpusDict],
    output_path: Optional[str] = None,
    conditions: Optional[List[Tuple[int, int, int]]] = None,
    envs: Optional[List[str]] = None,
    skip_duplicate_items: bool = False,
) -> List[str]:
    """Create a new corpus from the original one that contains the
    subsentences up to the position of the NPI.

    Parameters
    ----------
    orig_corpus : str | CorpusDict
        Either the path to the original corpus, or a CorpusDict that
        has been created using `preproc_warstadt`.
    output_path : str, optional
        Path to the output file that will be created in .tsv format.
        If not provided the corpus won't be written to disk.
    conditions : List[Tuple[int, int, int]], optional
        List of corpus item conditions (licensor, scope, npi_present).
        If not provided the correct NPI cases (1, 1, 1) will be used.
    envs : List[str], optional
        List of of licensing environments that should be used.
    skip_duplicate_items : bool
        Some corpus items only differ in their post-NPI content, and
        will lead to equivalent results on a downstream task. Defaults
        to False.

    Returns
    -------
    corpus : List[str]
        List of strings representing each corpus item. Note that the
        first line of the list contains the .tsv header.
    """
    if envs is None:
        envs = ENVS
    if conditions is None:
        conditions = [(1, 1, 1)]

    if isinstance(orig_corpus, str):
        id2items = preproc_warstadt(orig_corpus)
    else:
        id2items = orig_corpus

    sens_seen = set()
    corpus = [
        "\t".join(
            [
                "orig_sen_idx",
                "condition",
                "sen",
                "counter_sen",
                "token",
                "full_npi",
                "env",
                "labels",
            ]
        )
    ]

    for idx, items in id2items.items():
        if items[1, 1, 1]["env"] not in envs:
            continue

        for condition in conditions:
            licensor, scope, npi_present = condition
            item = items[condition]

            # Flip licensor bool to create an item of the opposite licensing polarity.
            counter_condition = (int(not licensor), scope, npi_present)
            counter_item = items[counter_condition]

            full_npi = item["npi"]
            # For multi-phrase NPIs (at all, in years) we are interested in the final token
            true_npi = full_npi.split()[-1]
            sen = " ".join(item["sen"])
            counter_sen = " ".join(counter_item["sen"])

            # Index of start of NPI phrase, to which we add the index of the final token
            start_idx = sen.index(f" {full_npi} ") + 1
            npi_idx = start_idx + sen[start_idx:].index(true_npi)
            start_idx = counter_sen.index(f" {full_npi} ") + 1
            counter_npi_idx = start_idx + counter_sen[start_idx:].index(true_npi)

            monotonicity = "downward" if licensor == 1 else "upward"

            if sen[:npi_idx] + full_npi in sens_seen and skip_duplicate_items:
                continue

            sens_seen.add(sen[:npi_idx] + full_npi)

            corpus.append(
                "\t".join(
                    (
                        str(idx),
                        str(condition),
                        sen[:npi_idx],
                        counter_sen[:counter_npi_idx],
                        true_npi,
                        full_npi,
                        item["env"],
                        monotonicity,
                    )
                )
            )

    if output_path is not None:
        with open(output_path, "w") as f:
            f.write("\n".join(corpus))

    return corpus


if __name__ == "__main__":
    new_corpus = create_downstream_corpus(
        "../../../lm_data/corpora/downstream/warstadt/npi_data_all_environments.tsv",
        output_path="../../../lm_data/corpora/npi/lc_detection_binary_NEW.tsv",
        conditions=[(1, 1, 1), (0, 1, 1)],
        skip_duplicate_items=True,
    )
    print(len(new_corpus))



from typing import List

from torch import Tensor

from diagnnose.models import LanguageModel
from diagnnose.typedefs.activations import ActivationDict, ActivationName


class RecurrentLM(LanguageModel):
    """ Abstract class for RNN LM with intermediate activations """

    is_recurrent: bool = True
    forget_offset: int = 0
    ih_concat_order: List[str] = ["h", "i"]
    split_order: List[str]
    use_char_embs: bool = False
    init_states: ActivationDict = {}

    @property
    def num_layers(self) -> int:
        return max(layer for layer, _name in self.sizes) + 1

    @property
    def top_layer(self) -> int:
        return self.num_layers - 1

    @property
    def output_size(self) -> int:
        return self.sizes[self.top_layer, "hx"]

    def init_hidden(self, batch_size: int) -> ActivationDict:
        """Creates a batch of initial states.

        Parameters
        ----------
        batch_size : int
            Size of batch for which states are created.

        Returns
        -------
        init_states : ActivationTensors
            Dictionary mapping hidden and cell state to init tensors.
        """
        batch_init_states: ActivationDict = {}

        for layer in range(self.num_layers):
            for hc in ["hx", "cx"]:
                # Shape: (batch_size, nhid)
                batched_state = self.init_states[layer, hc].repeat(batch_size, 1)
                batch_init_states[layer, hc] = batched_state

        return batch_init_states

    def final_hidden(self, hidden: ActivationDict) -> Tensor:
        """Returns the final hidden state.

        Parameters
        ----------
        hidden : ActivationTensors
            Dictionary of extracted activations.

        Returns
        -------
        final_hidden : Tensor
            Tensor of the final hidden state.
        """
        return hidden[self.top_layer, "hx"].squeeze()

    def nhid(self, activation_name: ActivationName) -> int:
        """Returns number of hidden units for a (layer, name) tuple.

        If `name` != emb/hx/cx returns the size of (layer, `cx`).
        """
        layer, name = activation_name

        return self.sizes.get((layer, name), self.sizes[layer, "cx"])



from typing import Optional

from transformers import PreTrainedTokenizer

from diagnnose.config.config_dict import create_config_dict
from diagnnose.corpus import Corpus
from diagnnose.extract import Extractor
from diagnnose.models import LanguageModel
from diagnnose.models.import_model import import_model
from diagnnose.tokenizer.create import create_tokenizer


class Pipeline:
    def __init__(self):
        config_dict = create_config_dict()

        self.tokenizer: Optional[PreTrainedTokenizer] = None
        self.corpus: Optional[Corpus] = None
        self.model: Optional[LanguageModel] = None
        self.extractor: Optional[Extractor] = None

        if "tokenizer" in config_dict:
            self.tokenizer: Optional[PreTrainedTokenizer] = create_tokenizer(
                **config_dict["tokenizer"]
            )
        if "corpus" in config_dict:
            self.corpus = Corpus.create(
                tokenizer=self.tokenizer, **config_dict["corpus"]
            )
        if "model" in config_dict:
            self.model = import_model(config_dict)
        if "extract" in config_dict:
            self.extractor = Extractor(
                self.model, self.corpus, **config_dict["extract"]
            )

    def run(self):
        if self.extractor is not None:
            self.extractor.extract()



from functools import reduce
from typing import Iterable, List

from torchtext.data import Example

from diagnnose.typedefs.activations import SelectionFunc


def return_all(_w_idx: int, _item: Example) -> bool:
    """ Always returns True for every token. """
    return True


def final_token(sen_column: str = "sen") -> SelectionFunc:
    """Only returns the final token of a sentence.

    Wrapper allows a different ``sen_column`` to be set, that indicates
    the ``sen`` attribute of a corpus item that is being processed.
    """

    def selection_func(w_idx: int, item: Example) -> bool:
        sen = getattr(item, sen_column)

        return w_idx == (len(sen) - 1)

    return selection_func


def final_sen_token(w_idx: int, item: Example) -> bool:
    """ Only returns the final token of a sentence. """
    sen = getattr(item, "sen")

    return w_idx == (len(sen) - 1)


def only_mask_token(mask_token: str, sen_column: str = "sen") -> SelectionFunc:
    def selection_func(w_idx: int, item: Example) -> bool:
        sen = getattr(item, sen_column)

        return sen[w_idx] == mask_token

    return selection_func


def first_n(n: int) -> SelectionFunc:
    """Wrapper that creates a selection_func that only returns True for
    the first `n` items of a corpus.
    """

    def selection_func(_w_idx: int, item: Example) -> bool:
        return item.sen_idx < n

    return selection_func


def nth_token(n: int) -> SelectionFunc:
    """Wrapper that creates a selection_func that only returns True for
    the `n^{th}` token of a sentence.
    """

    def selection_func(w_idx: int, _item: Example) -> bool:
        return w_idx == n

    return selection_func


def in_sen_ids(sen_ids: List[int]) -> SelectionFunc:
    """Wrapper that creates a selection_func that only returns True for
    a `sen_id` if it is part of the provided list of `sen_ids`.
    """

    def selection_func(_w_idx: int, item: Example) -> bool:
        return item.sen_idx in sen_ids

    return selection_func


# Higher-order boolean selection_func logic
def intersection(selection_funcs: Iterable[SelectionFunc]) -> SelectionFunc:
    """ Returns the intersection of an iterable of selection_funcs. """

    def selection_func(w_idx: int, item: Example) -> bool:
        return reduce(
            lambda out, func: out and func(w_idx, item), selection_funcs, True
        )

    return selection_func


def union(selection_funcs: Iterable[SelectionFunc]) -> SelectionFunc:
    """ Returns the union of an iterable of selection_funcs. """

    def selection_func(w_idx: int, item: Example) -> bool:
        return reduce(
            lambda out, func: out or func(w_idx, item), selection_funcs, False
        )

    return selection_func


def negate(selection_func: SelectionFunc) -> SelectionFunc:
    """ Returns the negation of a selection_func. """

    def neg_selection_func(w_idx: int, item: Example) -> bool:
        return not selection_func(w_idx, item)

    return neg_selection_func



import glob
import os
from typing import Dict, List, Optional, Sequence, Tuple

from torchtext.data import Example, Field

from diagnnose.corpus import Corpus
from diagnnose.typedefs.syntax import SyntaxEvalCorpora
from diagnnose.utils.pickle import load_pickle

from ..task import SyntaxEvalTask


class MarvinTask(SyntaxEvalTask):
    def initialize(
        self, path: str, subtasks: Optional[List[str]] = None
    ) -> SyntaxEvalCorpora:
        """Performs the initialization for the tasks of
        Marvin & Linzen (2018)

        Arxiv link: https://arxiv.org/pdf/1808.09031.pdf

        Repo: https://github.com/BeckyMarvin/LM_syneval

        Parameters
        ----------
        path : str
            Path to directory containing the Marvin datasets that can be
            found in the github repo.
        subtasks : List[str], optional
            The downstream tasks that will be tested. If not provided this
            will default to the full set of conditions.

        Returns
        -------
        corpora : Dict[str, Corpus]
            Dictionary mapping a subtask to a Corpus.
        """
        subtask_paths = glob.glob(os.path.join(path, "*.pickle"))
        all_subtasks = [os.path.basename(path).split(".")[0] for path in subtask_paths]
        subtask_to_path = dict(zip(all_subtasks, subtask_paths))

        subtasks: List[str] = subtasks or all_subtasks

        corpora: SyntaxEvalCorpora = {}

        for subtask in subtasks:
            subtask_path = subtask_to_path[subtask]
            subtask_corpora: Dict[str, Corpus] = self._initialize_subtask(
                subtask, subtask_path
            )

            corpora[subtask] = subtask_corpora

        return corpora

    def _initialize_subtask(self, subtask: str, corpus_path: str) -> Dict[str, Corpus]:
        corpus_dict: Dict[str, List[Sequence[str]]] = load_pickle(corpus_path)

        if "npi" in subtask:
            header = ["sen", "counter_sen", "token"]
            tokenize_columns = ["sen", "counter_sen"]
        else:
            header = ["sen", "token", "counter_token"]
            tokenize_columns = ["sen"]

        fields = Corpus.create_fields(
            header, tokenize_columns=tokenize_columns, tokenizer=self.tokenizer
        )
        subtask_corpora: Dict[str, Corpus] = {}

        for condition, sens in corpus_dict.items():
            examples = self._create_examples(subtask, sens, fields)

            corpus = Corpus(examples, fields)

            subtask_corpora[condition] = corpus

        return subtask_corpora

    def _create_examples(
        self, subtask: str, sens: List[Sequence[str]], fields: List[Tuple[str, Field]]
    ):
        if "npi" in subtask:
            return self._create_npi_examples(sens, fields)

        return self._create_sva_examples(sens, fields)

    @staticmethod
    def _create_sva_examples(
        sens: List[Sequence[str]], fields: List[Tuple[str, Field]]
    ) -> List[Example]:
        examples = []

        for s1, s2 in sens:
            s1, s2 = s1.split(), s2.split()

            # Locate index of verb as first point where correct and incorrect sentence differ.
            verb_index = 0
            for w1, w2 in zip(s1, s2):
                if w1 != w2:
                    break
                verb_index += 1

            subsen = s1[:verb_index]
            verb = s1[verb_index]
            wrong_verb = s2[verb_index]
            ex = Example.fromlist([subsen, verb, wrong_verb], fields)
            examples.append(ex)

        return examples

    @staticmethod
    def _create_npi_examples(
        sens: List[Sequence[str]], fields: List[Tuple[str, Field]]
    ) -> List[Example]:
        examples = []

        for s1, s2, _ in sens:
            s1, s2 = s1.split(), s2.split()

            npi = "ever"
            ever_idx = s1.index(npi)
            sen = " ".join(s1[:ever_idx])
            counter_sen = " ".join(s2[:ever_idx])

            ex = Example.fromlist([sen, counter_sen, npi], fields)
            examples.append(ex)

        return examples



from typing import Iterable, Optional, Sized

from torch import Tensor, long

from diagnnose.typedefs.activations import ActivationIndex


def activation_index_to_iterable(
    activation_index: ActivationIndex, stop_index: Optional[int] = None
) -> Iterable:
    """ Transforms an activation index into an iterable object. """
    if isinstance(activation_index, Tensor):
        activation_index = activation_index.to(long)

    if isinstance(activation_index, Iterable):
        return activation_index

    if isinstance(activation_index, int):
        return [activation_index]

    if isinstance(activation_index, slice):
        stop_index = activation_index.stop or stop_index
        assert stop_index is not None, "Stop index of slice should be provided"
        return range(
            activation_index.start or 0, stop_index, activation_index.step or 1
        )

    raise ValueError(
        f"Activation index of incorrect type: {type(activation_index)}, "
        f"should be one of {{int, List[int], np.ndarray or torch.Tensor}}"
    )


def activation_index_len(activation_index: ActivationIndex) -> int:
    """ Returns the number of items in an activation index. """
    activation_iterable = activation_index_to_iterable(activation_index)
    if isinstance(activation_iterable, Sized):
        return len(activation_iterable)

    raise ValueError(
        f"Activation index of incorrect type: {type(activation_index)}, "
        f"should be one of {{int, List[int], np.ndarray or torch.Tensor}}"
    )



from typing import BinaryIO, Callable, Dict, List, Tuple, Union

from torch import Tensor
from torchtext.data import Example

# ACTIVATION DICTS
ActivationName = Tuple[int, str]  # (layer, name)
ActivationNames = List[ActivationName]

ActivationDict = Dict[ActivationName, Tensor]

# LM's layer sizes: (layer, name) -> size
SizeDict = Dict[ActivationName, int]

# Maps a layer index to a tensor
LayeredTensors = Dict[int, Tensor]


# EXTRACTION
ActivationFiles = Dict[ActivationName, BinaryIO]

# token index, corpus item -> bool
SelectionFunc = Callable[[int, Example], bool]

# [(start, stop)]
ActivationRanges = List[Tuple[int, int]]

RemoveCallback = Callable[[], None]


# INDEXING
# Activation indexing, as done in ActivationReader
ActivationIndex = Union[int, slice, List[int], Tensor]

ActivationKey = Union[ActivationIndex, Tuple[ActivationIndex, ActivationName]]



from typing import Any, Callable, Iterable, List, Optional, Tuple, Union
from warnings import warn

import torch
from torch import Tensor

from . import utils


class ShapleyTensor:
    """A ShapleyTensor wraps a torch Tensor. It allows the tensor to
    be decomposed into a sum of tensors, that each define the
    contribution of a feature to the tensor.

    ShapleyTensors can be passed to any type of torch model. For each
    operation in the model the intermediate Shapley values are
    calculated for the list of contributions. This is done using
    `__torch_function__`, that allows to override tensor operations.

    Parameters
    ----------
    data : Tensor
        Input tensor that is decomposed into a sum of contributions.
    contributions : List[Tensor]
        List of contributions that should sum up to `data`.
    shapley_factors : List[Tuple[List[int], int]], optional
        Shapley factors that are calculated with `calc_shapley_factors`.
        To prevent unnecessary compute these factors are passed on to
        subsequent ShapleyTensors.
    num_samples : int, optional
        Number of feature permutation samples. Increasing the number of
        samples will reduce the variance of the approximation. If not
        provided the exact Shapley values will be computed.
    validate : bool, optional
        Toggle to validate at each step whether `contributions` still
        sums up to `data`. Defaults to False.
    baseline_partition : int, optional
        Index of the contribution partition to which the baseline fn(0)
        will be added. If we do not add this baseline the contributions
        won't sum up to the full output. Defaults to 0.
    """

    def __init__(
        self,
        data: Tensor,
        contributions: Optional[List[Tensor]] = None,
        shapley_factors: Optional[List[Tuple[List[int], int]]] = None,
        num_samples: Optional[int] = None,
        validate: bool = False,
        baseline_partition: int = 0,
    ):
        if not utils.MONKEY_PATCH_PERFORMED:
            utils.monkey_patch()
            utils.MONKEY_PATCH_PERFORMED = True

        self.data = data
        self.contributions = contributions or []
        self.shapley_factors = shapley_factors
        self.num_samples = num_samples
        self.validate = validate
        self.baseline_partition = baseline_partition

        self.current_fn: Optional[str] = None
        self.new_data: Optional[Union[Tensor, Iterable[Tensor]]] = None

        if (
            len(self.contributions) > 0
            and shapley_factors is None
            and num_samples is None
        ):
            self.shapley_factors = utils.calc_shapley_factors(self.num_features)

    def __torch_function__(self, fn, _types, args=(), kwargs=None):
        self.current_fn = fn.__name__

        kwargs = kwargs or {}

        self.new_data = fn(*map(utils.unwrap, args), **kwargs)

        new_contributions = self._calc_contributions(fn, *args, **kwargs)

        return self._pack_output(self.new_data, new_contributions)

    @property
    def num_features(self) -> int:
        return len(self.contributions)

    def size(self, *args, **kwargs):
        return self.data.size(*args, **kwargs)

    def __iter__(self):
        """Allows a ShapleyTensor to be unpacked directly as:

        .. code-block:: python

            data, contributions = shapley_tensor
        """
        yield from [self.data, self.contributions]

    def __len__(self):
        return len(self.data)

    def __getattr__(self, item: str) -> Any:
        """
        Handles torch methods that are called on a tensor itself, like
        ``tensor.add(*args)`` or ``tensor.view(*args)``.
        """
        attr = getattr(self.data, item)

        if isinstance(attr, Callable):

            def attr_wrapper(*args, **kwargs):
                torch_fn = getattr(torch, attr.__name__, None)

                if isinstance(torch_fn, Callable):
                    # Captures function calls like tensor.transpose(*args), and returns them
                    # like torch.transpose(tensor, *args), so the function call can be captured
                    # using __torch_function__.
                    if attr.__name__ == "reshape":
                        return torch_fn(self, args, **kwargs)

                    return torch_fn(self, *args, **kwargs)
                else:
                    # Captures tensor functions that don't exist as a stand-alone torch method,
                    # such as tensor.view(*args). Applies the same function to all contributions.
                    output = attr(*args, **kwargs)
                    contributions = [
                        getattr(contribution, item)(*args, **kwargs)
                        for contribution in self.contributions
                    ]

                    return self._pack_output(output, contributions)

            return attr_wrapper
        else:
            return attr

    def __getitem__(self, index):
        if isinstance(index, ShapleyTensor):
            data = self.data[index.data]
            contributions = [self.data[c] for c in index.contributions]
            self.validate = self.validate or index.validate
        else:
            data = self.data[index]
            contributions = [contribution[index] for contribution in self.contributions]

        # We return type(self) to allow a subclass that derives from ShapleyTensor to be preserved.
        tensor_type = type(self)

        return tensor_type(
            data,
            contributions=contributions,
            shapley_factors=self.shapley_factors,
            num_samples=self.num_samples,
            validate=self.validate,
            baseline_partition=self.baseline_partition,
        )

    def __setitem__(self, index, value):
        self.data[index] = value.data

        # We pad the current contributions if the value that is set contains more contributions
        # than the current ShapleyTensor.
        if len(self.contributions) < len(value.contributions):
            extra_contributions = len(value.contributions) - len(self.contributions)
            for _ in range(extra_contributions):
                self.contributions.append(torch.zeros_like(self.data))

        for c_idx, contribution in enumerate(self.contributions):
            contribution[index] = value.contributions[c_idx]

    def _validate_contributions(self, data, contributions) -> None:
        """ Asserts whether the contributions sum up to the full tensor. """
        diff = (data - sum(contributions)).float()
        mean_diff = torch.mean(diff)
        max_diff = torch.max(torch.abs(diff))
        if not torch.allclose(data, sum(contributions), rtol=1e-3, atol=1e-3):
            warn(
                f"Contributions don't sum up to the provided tensor, with a mean difference of "
                f"{mean_diff:.3E} and a max difference of {max_diff:.3E}. "
                f"Current function is: {self.current_fn}"
            )

    def _pack_output(
        self,
        new_data: Union[Tensor, Iterable[Tensor]],
        new_contributions: Union[List[Tensor], Iterable[List[Tensor]]],
    ) -> Any:
        """Packs the output and its corresponding contributions into a
        new ShapleyTensor.

        If the output is an iterable (e.g. with a .split operation) the
        type structure of the output is preserved.
        """
        if isinstance(new_data, torch.Tensor):
            tensor_type = type(self)

            if self.validate and len(new_contributions) > 0:
                self._validate_contributions(new_data, new_contributions)

            return tensor_type(
                new_data,
                contributions=new_contributions,
                shapley_factors=self.shapley_factors,
                num_samples=self.num_samples,
                validate=self.validate,
                baseline_partition=self.baseline_partition,
            )
        elif self.current_fn == "_pack_padded_sequence":
            new_contributions = [c[0] for c in new_contributions]

            return self._pack_output(new_data[0], new_contributions), new_data[1]
        elif isinstance(new_data, (list, tuple)):
            iterable_type = type(new_data)

            if len(new_contributions) == 0:
                return iterable_type(self._pack_output(item, []) for item in new_data)

            return iterable_type(
                self._pack_output(item, new_contributions[idx])
                for idx, item in enumerate(new_data)
            )

        return new_data

    def _calc_contributions(self, fn, *args, **kwargs) -> List[Tensor]:
        """
        Some methods have custom behaviour for how the output is
        decomposed into a new set of contributions.
        """
        if self.num_features == 0:
            return []
        elif hasattr(self, f"{fn.__name__}_contributions"):
            fn = getattr(self, f"{fn.__name__}_contributions")
            return fn(*args, **kwargs)
        elif fn.__name__ in [
            "squeeze",
            "unsqueeze",
            "index_select",
            "_pack_padded_sequence",
        ]:
            old_contributions = args[0].contributions
            return [fn(c, *args[1:], **kwargs) for c in old_contributions]

        return self._calc_shapley_contributions(fn, *args, **kwargs)

    def _calc_shapley_contributions(self, fn, *args, **kwargs) -> List[Tensor]:
        """ Calculates the Shapley decomposition of the current fn. """
        assert isinstance(
            self.new_data, Tensor
        ), f"Current operation {self.current_fn} is not supported for Shapley calculation"

        if self.num_samples is None:
            return utils.calc_exact_shapley_values(
                fn,
                self.num_features,
                self.shapley_factors,
                self.new_data,
                self.baseline_partition,
                *args,
                **kwargs,
            )
        else:
            return utils.calc_sample_shapley_values(
                fn,
                self.num_features,
                self.num_samples,
                self.new_data,
                self.baseline_partition,
                *args,
                **kwargs,
            )

    def cat_contributions(self, *args, **kwargs):
        def _pad_contributions(arg: Union[Tensor, ShapleyTensor]):
            """A non-ShapleyTensor only contributes to the baseline
            partition, and is padded with 0s.
            """
            if isinstance(arg, ShapleyTensor):
                return arg.contributions
            elif isinstance(arg, Tensor):
                padded_contribution = self.num_features * [torch.zeros_like(arg)]
                padded_contribution[self.baseline_partition] = arg

                return padded_contribution
            else:
                raise TypeError

        padded_contributions = [_pad_contributions(tensor) for tensor in args[0]]

        contributions = [
            torch.cat(
                [contribution[idx] for contribution in padded_contributions], **kwargs
            )
            for idx in range(self.num_features)
        ]

        return contributions

    @staticmethod
    def split_contributions(*args, **kwargs):
        shapley_tensor, split_size_or_sections = args

        raw_splits = [
            torch.split(contribution, split_size_or_sections, **kwargs)
            for contribution in shapley_tensor.contributions
        ]

        num_splits = len(raw_splits[0])

        contributions = [
            [split[idx] for split in raw_splits] for idx in range(num_splits)
        ]

        return contributions

    def dropout_contributions(self, *args, **kwargs):
        """In principle dropout should be disabled when calculating
        Shapley contributions, but we should still take care of it.

        We determine the dropout mask by looking at the difference
        between the new output data and the input.
        """
        dropout_mask = self.new_data != args[0].data

        contributions = args[0].contributions
        for contribution in contributions:
            contribution[dropout_mask] = 0.0

        return contributions

    def dropout2d_contributions(self, *args, **kwargs):
        return self.dropout_contributions(*args, **kwargs)

    def dropout3d_contributions(self, *args, **kwargs):
        return self.dropout_contributions(*args, **kwargs)

    def add_contributions(self, *args, **kwargs):
        """ Non-ShapleyTensors are added to the baseline partition. """
        input_, other = args

        if not isinstance(input_, ShapleyTensor):
            contributions = other.contributions
            contributions[self.baseline_partition] += input_
        elif not isinstance(other, ShapleyTensor):
            contributions = input_.contributions
            contributions[self.baseline_partition] += other
        else:
            contributions = [
                torch.add(con1, con2, **kwargs)
                for con1, con2 in zip(input_.contributions, other.contributions)
            ]

        return contributions

    def mul_contributions(self, *args, **kwargs):
        input_, other = args
        if isinstance(input_, torch.Tensor):
            contributions = [
                torch.mul(input_, contribution, **kwargs)
                for contribution in other.contributions
            ]
        elif isinstance(other, torch.Tensor):
            contributions = [
                torch.mul(contribution, other, **kwargs)
                for contribution in input_.contributions
            ]
        else:
            contributions = self._calc_shapley_contributions(torch.mul, *args, **kwargs)

        return contributions

    def matmul_contributions(self, *args, **kwargs):
        input_, other = args
        if isinstance(input_, torch.Tensor):
            contributions = [
                torch.matmul(input_, contribution, **kwargs)
                for contribution in other.contributions
            ]
        elif isinstance(other, torch.Tensor):
            contributions = [
                torch.matmul(contribution, other, **kwargs)
                for contribution in input_.contributions
            ]
        else:
            contributions = self._calc_shapley_contributions(
                torch.matmul, *args, **kwargs
            )

        return contributions

    def __add__(self, other):
        return torch.add(self, other)

    def __radd__(self, other):
        return torch.add(other, self)

    def __sub__(self, other):
        return torch.sub(self, other)

    def __rsub__(self, other):
        return torch.sub(other, self)

    def __matmul__(self, other):
        return torch.matmul(self, other)

    def __rmatmul__(self, other):
        return torch.matmul(other, self)

    def __mul__(self, other):
        return torch.mul(self, other)

    def __rmul__(self, other):
        return torch.mul(other, self)

    def __pow__(self, other):
        return torch.pow(self, other)

    def __div__(self, other):
        return torch.div(self, other)

    def __rdiv__(self, other):
        return torch.div(other, self)

    def __mod__(self, other):
        return torch.fmod(self, other)

    def __truediv__(self, other):
        return torch.true_divide(self, other)

    def __floordiv__(self, other):
        return torch.div(self, other).floor()

    def __rfloordiv__(self, other):
        return torch.div(other, self).floor()

    def __abs__(self):
        return torch.abs(self)

    def __and__(self, other):
        return torch.logical_and(self, other)

    def __ge__(self, other):
        return torch.ge(self, other)

    def __gt__(self, other):
        return torch.gt(self, other)

    def __invert__(self):
        return torch.logical_not(self)

    def __le__(self, other):
        return torch.le(self, other)

    def __lt__(self, other):
        return torch.lt(self, other)

    def __ne__(self, other):
        return torch.ne(self, other)

    def __neg__(self):
        return torch.neg(self)

    def __or__(self, other):
        return torch.logical_or(self, other)

    def __xor__(self, other):
        return torch.logical_xor(self, other)



import os
from itertools import product
from typing import Optional

import torch
from transformers import PreTrainedTokenizer

from diagnnose.activations.selection_funcs import final_sen_token
from diagnnose.corpus import Corpus
from diagnnose.extract import Extractor
from diagnnose.typedefs.activations import ActivationDict, ActivationNames
from diagnnose.utils import __file__ as diagnnose_utils_init
from diagnnose.utils.misc import suppress_print
from diagnnose.utils.pickle import load_pickle

from .recurrent_lm import RecurrentLM


def set_init_states(
    model: RecurrentLM,
    pickle_path: Optional[str] = None,
    corpus_path: Optional[str] = None,
    use_default: bool = False,
    tokenizer: Optional[PreTrainedTokenizer] = None,
    save_init_states_to: Optional[str] = None,
) -> None:
    """Set up the initial LM states.

    If no path is provided 0-valued embeddings will be used.
    Note that the loaded init should provide tensors for `hx`
    and `cx` in all layers of the LM.

    Note that `init_states_pickle` takes precedence over
    `init_states_corpus` in case both are provided.

    Parameters
    ----------
    model : RecurrentLM
        A recurrent language model for which the initial states are set.
    pickle_path : str, optional
        Path to pickled file with initial lstm states. If not
        provided zero-valued init states will be created.
    corpus_path : str, optional
        Path to corpus of which the final hidden state will be used
        as initial states.
    use_default : bool
        Toggle to use the default initial sentence `. <eos>`.
    tokenizer : PreTrainedTokenizer, optional
        Tokenizer that must be provided when creating the init
        states from a corpus.
    save_init_states_to : str, optional
        Path to which the newly computed init_states will be saved.
        If not provided these states won't be dumped.

    Returns
    -------
    init_states : ActivationTensors
        ActivationTensors containing the init states for each layer.
    """
    if use_default:
        diagnnose_utils_dir = os.path.dirname(diagnnose_utils_init)
        corpus_path = os.path.join(diagnnose_utils_dir, "init_sentence.txt")

    if pickle_path is not None:
        init_states = _create_init_states_from_pickle(model, pickle_path)
    elif corpus_path is not None:
        init_states = _create_init_states_from_corpus(
            model, corpus_path, tokenizer, save_init_states_to
        )
    else:
        init_states = _create_zero_states(model)

    model.init_states = init_states


def _create_zero_states(model: RecurrentLM) -> ActivationDict:
    """Zero-initialized states if no init state is provided.

    Returns
    -------
    init_states : ActivationTensors
        Dictionary mapping (layer, name) tuple to zero-tensor.
    """
    init_states: ActivationDict = {
        a_name: torch.zeros((1, model.nhid(a_name)))
        for a_name in product(range(model.num_layers), ["cx", "hx"])
    }

    return init_states


@suppress_print
def _create_init_states_from_corpus(
    model: RecurrentLM,
    init_states_corpus: str,
    tokenizer: PreTrainedTokenizer,
    save_init_states_to: Optional[str] = None,
) -> ActivationDict:
    assert (
        tokenizer is not None
    ), "Tokenizer must be provided when creating init states from corpus"

    corpus: Corpus = Corpus.create(init_states_corpus, tokenizer=tokenizer)

    activation_names: ActivationNames = [
        (layer, name) for layer in range(model.num_layers) for name in ["hx", "cx"]
    ]

    extractor = Extractor(
        model,
        corpus,
        activation_names,
        activations_dir=save_init_states_to,
        selection_func=final_sen_token,
    )
    init_states = extractor.extract().activation_dict

    return init_states


def _create_init_states_from_pickle(
    model: RecurrentLM, pickle_path: str
) -> ActivationDict:
    init_states: ActivationDict = load_pickle(pickle_path)

    _validate_init_states_from_pickle(model, init_states)

    return init_states


def _validate_init_states_from_pickle(
    model: RecurrentLM, init_states: ActivationDict
) -> None:
    num_init_layers = max(layer for layer, _name in init_states)
    assert num_init_layers == model.num_layers, "Number of initial layers not correct"

    for (layer, name), size in model.sizes.items():
        if name in ["hx", "cx"]:
            assert (
                layer,
                name,
            ) in init_states.keys(), (
                f"Activation {layer},{name} is not found in init states"
            )

            init_size = init_states[layer, name].size(1)
            assert init_size == size, (
                f"Initial activation size for {name} is incorrect: "
                f"{name}: {init_size}, should be {size}"
            )



"""Model wrappers define the behaviour of a specific model architecture.

These wrapper classes should inherit from :class:`LanguageModel`, or one
of its subclasses, and adhere to the signature of the ``forward`` method
that is defined for it.
"""

from .awd_lstm import AWDLSTM
from .forward_lstm import ForwardLSTM
from .google_lm import GoogleLM



from typing import Any, Dict, List, Optional, Type

from transformers import PreTrainedTokenizer

from diagnnose.models import LanguageModel
from diagnnose.typedefs.syntax import ResultsDict

from .task import SyntaxEvalTask
from .tasks import LakretzTask, LinzenTask, MarvinTask, WarstadtTask, WinobiasTask

task_constructors: Dict[str, Type[SyntaxEvalTask]] = {
    "lakretz": LakretzTask,
    "linzen": LinzenTask,
    "marvin": MarvinTask,
    "warstadt": WarstadtTask,
    "winobias": WinobiasTask,
}


class SyntacticEvaluator:
    """Suite that runs multiple syntactic evaluation tasks on a LM.

    Tasks can be run on already extracted activations, or on a new LM
    for which new activations will be extracted.

    Initialisation is performed separately from the tasks themselves,
    in order to allow multiple LMs to be ran on the same set of tasks.

    Parameters
    ----------
    tokenizer : PreTrainedTokenizer
        Tokenizer that converts tokens to their index within the LM.
    config : Dict[str, Any]
        Dictionary mapping a task name (`lakretz`, `linzen`, `marvin`,
        `warstadt`, or `winobias`) to its configuration (`path`,
        `tasks`, and `task_activations`). `path` points to the corpus
        folder of the task, `tasks` is an optional list of subtasks, and
        `task_activations` an optional path to the folder containing
        the model activations.
    """

    def __init__(
        self,
        model: LanguageModel,
        tokenizer: PreTrainedTokenizer,
        config: Dict[str, Any],
        ignore_unk: bool = False,
        use_full_model_probs: bool = True,
        tasks: Optional[List[str]] = None,
    ) -> None:
        self.tasks: Dict[str, SyntaxEvalTask] = {}

        print("Initializing syntactic evaluation tasks...")

        for task_name in tasks or config.keys():
            constructor = task_constructors.get(task_name, SyntaxEvalTask)

            self.tasks[task_name] = constructor(
                model,
                tokenizer,
                ignore_unk=ignore_unk,
                use_full_model_probs=use_full_model_probs,
                **config[task_name],
            )

        print("Syntactic evaluation task initialization finished")

    def run(self) -> Dict[str, Any]:
        results: Dict[str, ResultsDict] = {}

        for task_name, task in self.tasks.items():
            print(f"\n--=={task_name.upper()}==--")

            results[task_name] = task.run()

        return results



import itertools
from functools import wraps
from math import factorial
from typing import Any, Callable, Iterable, List, Optional, Tuple

import torch
from torch import Tensor

try:
    # torch 1.5
    from torch._overrides import handle_torch_function, has_torch_function
except ModuleNotFoundError:
    # torch >1.5
    from torch.overrides import handle_torch_function, has_torch_function

MONKEY_PATCH_PERFORMED = False


def monkey_patch():
    """Not all torch functions correctly implement ``__torch_function__``
    yet (i.e. in torch v1.5), as is discussed here:
    https://github.com/pytorch/pytorch/issues/34294

    We override the ``__torch_function__`` behaviour for ``torch.cat``,
    ``torch.stack``, ``Tensor.expand_as``, and ``Tensor.type_as``.
    """
    torch.cat = _monkey_patch_fn(torch.cat)
    torch.stack = _monkey_patch_fn(torch.stack)
    Tensor.expand_as = _monkey_patch_tensor(Tensor.expand_as)
    Tensor.type_as = _monkey_patch_tensor(Tensor.type_as)


def _monkey_patch_fn(original_fn):
    @wraps(original_fn)
    def fn(tensors, dim=0, out=None):
        if not torch.jit.is_scripting():
            if any(type(t) is not Tensor for t in tensors) and has_torch_function(
                tensors
            ):
                return handle_torch_function(fn, tensors, tensors, dim=dim, out=out)
        return original_fn(tensors, dim=dim, out=out)

    return fn


def _monkey_patch_tensor(original_fn):
    @wraps(original_fn)
    def fn(self, other):
        if isinstance(other, Tensor):
            return original_fn(self, other)
        return original_fn(self, other.data)

    return fn


def unwrap(args: Any, attr: str = "data", coalition: Optional[List[int]] = None) -> Any:
    """Unwraps a list of args that might contain ShapleyTensors.

    Can be used to retrieve: 1. The full tensor of each
    ShapleyTensor, 2. The list of contributions, or 3. The sum of
    contributions for a specific coalition.

    Unwrapping is performed recursively. Non-ShapleyTensors are left
    unchanged.

    Parameters
    ----------
    args : Any
        Either the full list of args, or an individual element of that
        list, as unwrapping is performed recursively.
    attr : str, optional
        The ShapleyTensor attribute that should be returned, either
        `data` or `contributions`.
    coalition : List[int], optional
        Optional list of coalition indices. If provided the
        contributions at the indices of the coalition are summed up and
        returned, instead of the full list of contributions.
    """
    if hasattr(args, attr):
        args_attr = getattr(args, attr)
        if coalition is not None:
            return sum_contributions(args_attr, coalition)
        return args_attr
    elif isinstance(args, (Tensor, str)):
        return args
    elif isinstance(args, list):
        return [unwrap(arg, attr, coalition) for arg in args]
    elif isinstance(args, tuple):
        return tuple(unwrap(arg, attr, coalition) for arg in args)

    return args


def sum_contributions(contributions: List[Tensor], coalition: List[int]) -> Tensor:
    """ Sums the contributions that are part of the provided coalition. """
    contributions_sum = sum([contributions[idx] for idx in coalition])
    if isinstance(contributions_sum, int):
        contributions_sum = torch.zeros_like(contributions[0])

    return contributions_sum


def calc_shapley_factors(num_features: int) -> List[Tuple[List[int], int]]:
    """Creates the normalization factors for each subset of features.

    These factors are based on the original Shapley formulation:
    https://en.wikipedia.org/wiki/Shapley_value

    If, for instance, we were to compute these factors for item
    :math:`a` in the set :math:`N = \{a, b, c\}`, we would pass
    :math:`|N|`. This returns the list
    :math:`[([], 2), ([0], 1), ([1], 1), ([0, 1], 2])]`. The first item
    of each tuple should be interpreted as the indices for the set
    :math:`N\setminus\{a\}: (0 \Rightarrow b, 1 \Rightarrow c)`, mapped
    to their factors: :math:`|ids|! \cdot (n - |ids|)!`.

    Parameters
    ----------
    num_features : int
        Number of features for which Shapley values will be computed.

    Returns
    -------
    shapley_factors : List[Tuple[List[int], int]]
        Dictionary mapping a tuple of indices to its corresponding
        normalization factor.
    """
    shapley_factors = []

    for i in range(num_features):
        factor = factorial(i) * factorial(num_features - i - 1)
        for pi in itertools.combinations(range(num_features - 1), i):
            shapley_factors.append((list(pi), factor))

    return shapley_factors


def perm_generator(num_features: int, num_samples: int) -> Iterable[List[int]]:
    """ Generator for feature index permutations. """
    for _ in range(num_samples):
        yield torch.randperm(num_features).tolist()


def calc_exact_shapley_values(
    fn: Callable,
    num_features: int,
    shapley_factors: List[Tuple[List[int], int]],
    new_data: Tensor,
    baseline_partition: int,
    *args,
    **kwargs,
) -> List[Tensor]:
    """Calculates the exact Shapley values for some function fn.

    Note that this procedure grows exponentially in the number of
    features, and should be handled with care.

    Parameters
    ----------
    fn : Callable
        Torch function for which the Shapley values will be computed.
    num_features : int
        Number of features for which contributions will be computed.
    shapley_factors : List[Tuple[List[int], int]]
        List of `Shapley factors` that is computed using
        ``calc_shapley_factors``.
    new_data : Tensor
        The output tensor that is currently being decomposed into
        its contributions. We pass this so we can instantiate the
        contribution tensors with correct shape beforehand.
    baseline_partition : int
        Index of the contribution partition to which the baseline fn(0)
        will be added. If we do not add this baseline the contributions
        won't sum up to the full output.
    """
    contributions = []

    for f_idx in range(num_features):
        other_ids = torch.tensor([i for i in range(num_features) if i != f_idx])

        contribution = torch.zeros_like(new_data)

        for coalition_ids, factor in shapley_factors:
            coalition = list(other_ids[coalition_ids])
            args_wo = unwrap(args, attr="contributions", coalition=coalition)

            args_with = unwrap(
                args, attr="contributions", coalition=(coalition + [f_idx])
            )

            contribution += factor * (fn(*args_with, **kwargs) - fn(*args_wo, **kwargs))

        contribution /= factorial(num_features)
        contributions.append(contribution)

    # Add baseline to default feature ([0]).
    zero_input_args = unwrap(args, attr="contributions", coalition=[])
    baseline = fn(*zero_input_args, **kwargs)
    contributions[baseline_partition] += baseline

    return contributions


def calc_sample_shapley_values(
    fn: Callable,
    num_features: int,
    num_samples: int,
    new_data: Tensor,
    baseline_partition: int,
    *args,
    **kwargs,
) -> List[Tensor]:
    """Calculates the approximate Shapley values for some function fn.

    This procedure is based on that of Castro et al. (2008), and
    approximates Shapley values in polynomial time.

    Parameters
    ----------
    fn : Callable
        Torch function for which the Shapley values will be computed.
    num_features : int
        Number of features for which contributions will be computed.
    num_samples : int
        Number of feature permutation samples. Increasing the number of
        samples will reduce the variance of the approximation.
    new_data : Tensor
        The output tensor that is currently being decomposed into
        its contributions. We pass this so we can instantiate the
        contribution tensors with correct shape beforehand.
    baseline_partition : int
        Index of the contribution partition to which the baseline fn(0)
        will be added. If we do not add this baseline the contributions
        won't sum up to the full output.
    """
    contributions = [torch.zeros_like(new_data) for _ in range(num_features)]

    generator = perm_generator(num_features, num_samples)

    zero_input_args = unwrap(args, attr="contributions", coalition=[])
    baseline = fn(*zero_input_args, **kwargs)

    for sample in generator:
        prev_value = baseline
        for sample_idx, feature_idx in enumerate(sample, start=1):
            coalition = sample[:sample_idx]
            coalition_args = unwrap(args, attr="contributions", coalition=coalition)

            new_value = fn(*coalition_args, **kwargs)
            contributions[feature_idx] += new_value - prev_value
            prev_value = new_value

    contributions = [c / num_samples for c in contributions]
    contributions[baseline_partition] += baseline

    return contributions



import os
from typing import Dict, List, Optional, Union

from transformers import AutoTokenizer, PreTrainedTokenizer

from .c2i import C2I
from .w2i import W2I


def token_to_index(path: str) -> Dict[str, int]:
    """Reads in a newline-separated file of tokenizer entries.

    Parameters
    ----------
    path : str
        Path to a vocabulary file.

    Returns
    -------
    w2i : Dict[str, int]
        Dictionary mapping a token string to its index.
    """
    with open(os.path.expanduser(path), encoding="ISO-8859-1") as f:
        w2i = {token.strip(): idx for idx, token in enumerate(f)}

    return w2i


def create_tokenizer(
    path: str, notify_unk: bool = False, cache_dir: Optional[str] = None
) -> PreTrainedTokenizer:
    """Creates a tokenizer from a path.

    A LSTM tokenizer is defined as a file with an entry at each line,
    and `path` should point towards that file.

    A Transformer tokenizer is defined by its model name, and is
    imported using the AutoTokenizer class.

    Parameters
    ----------
    path : str
        Either the path towards a vocabulary file, or the model name
        of a Huggingface Transformer.
    notify_unk : bool, optional
        Optional toggle to notify a user if a token is not present in
        the vocabulary of the tokenizer. Defaults to False.
    cache_dir : str, optional
        Cache directory for Huggingface tokenizers.

    Returns
    -------
    tokenizer : PreTrainedTokenizer
        The instantiated tokenizer that maps tokens to their indices.
    """
    if os.path.exists(os.path.expanduser(path)):
        # Word-based vocabulary, used by older LSTM models
        vocab = W2I(token_to_index(path), notify_unk=notify_unk)

        tokenizer = PreTrainedTokenizer()

        tokenizer.added_tokens_encoder = vocab
        tokenizer.added_tokens_decoder = {idx: w for w, idx in vocab.items()}
        tokenizer.vocab = tokenizer.added_tokens_encoder
        tokenizer.ids_to_tokens = tokenizer.added_tokens_decoder

        tokenizer.unk_token = vocab.unk_token
        tokenizer.eos_token = vocab.eos_token
        tokenizer.pad_token = vocab.pad_token

        tokenizer._tokenize = lambda s: s.split(" ")
        tokenizer._convert_token_to_id = lambda w: vocab[w]

        return tokenizer

    # Subword-based vocabulary, used by Transformer models
    tokenizer = AutoTokenizer.from_pretrained(path, cache_dir=cache_dir, use_fast=False)
    if hasattr(tokenizer, "encoder"):
        # GPT-2 & Roberta use a different attribute for the underlying vocab dictionary.
        encoder: Dict[str, int] = getattr(tokenizer, "encoder")
        tokenizer.vocab = W2I(encoder, unk_token=tokenizer.unk_token)
        tokenizer.ids_to_tokens = tokenizer.decoder

    if getattr(tokenizer, "pad_token", None) is None:
        tokenizer.pad_token = tokenizer.unk_token

    return tokenizer


def create_char_vocab(corpus_path: Union[str, List[str]]) -> C2I:
    return C2I(token_to_index(corpus_path))



from typing import List, Optional, Tuple, Union

from torchtext.data import Dataset, Example, Field, Pipeline, RawField
from transformers import PreTrainedTokenizer


class Corpus(Dataset):
    def __init__(
        self,
        examples: List[Example],
        fields: List[Tuple[str, Field]],
        create_pos_tags: bool = False,
        sen_column: str = "sen",
        labels_column: Optional[str] = None,
    ) -> None:
        super().__init__(examples, fields)

        self.sen_column = sen_column
        self.labels_column = labels_column

        self.tokenizer: Optional[PreTrainedTokenizer] = None
        if hasattr(self.fields[sen_column], "vocab"):
            self.tokenizer = self.fields[sen_column].vocab

        if "sen_idx" not in self.fields:
            self._attach_sen_ids()

        if create_pos_tags:
            self._create_pos_tags()

    @classmethod
    def create(
        cls,
        path: str,
        header: Optional[List[str]] = None,
        header_from_first_line: bool = False,
        to_lower: bool = False,
        sen_column: str = "sen",
        labels_column: Optional[str] = None,
        sep: str = "\t",
        tokenize_columns: Optional[List[str]] = None,
        convert_numerical: bool = False,
        create_pos_tags: bool = False,
        tokenizer: Optional[PreTrainedTokenizer] = None,
    ) -> "Corpus":
        raw_corpus = cls.create_raw_corpus(
            path, header_from_first_line=header_from_first_line, sep=sep
        )

        header = cls.create_header(
            header=header,
            header_from_first_line=header_from_first_line,
            corpus_path=path,
            sen_column=sen_column,
            sep=sep,
        )

        fields = cls.create_fields(
            header,
            to_lower=to_lower,
            sen_column=sen_column,
            tokenize_columns=tokenize_columns,
            convert_numerical=convert_numerical,
            tokenizer=tokenizer,
        )

        examples = cls.create_examples(raw_corpus, fields)

        return cls(
            examples,
            fields,
            create_pos_tags=create_pos_tags,
            sen_column=sen_column,
            labels_column=labels_column,
        )

    @staticmethod
    def create_raw_corpus(
        path: str, header_from_first_line: bool = False, sep: str = "\t"
    ) -> List[List[str]]:
        with open(path, encoding="utf8") as f:
            if header_from_first_line:
                next(f)
            raw_corpus = [line.strip().split(sep) for line in f]

        return raw_corpus

    @staticmethod
    def create_header(
        header: Optional[List[str]] = None,
        header_from_first_line: bool = False,
        corpus_path: Optional[str] = None,
        sen_column: str = "sen",
        sep: str = "\t",
    ) -> List[str]:
        if header is None:
            if header_from_first_line:
                with open(corpus_path) as f:
                    header = next(f).strip("\n").split(sep)
            else:
                header = [sen_column]

        assert sen_column in header, f"{sen_column} should be part of corpus_header!"

        return header

    @staticmethod
    def create_fields(
        header: List[str],
        to_lower: bool = False,
        sen_column: str = "sen",
        tokenize_columns: Optional[List[str]] = None,
        convert_numerical: bool = False,
        tokenizer: Optional[PreTrainedTokenizer] = None,
    ) -> List[Tuple[str, Field]]:
        tokenize_columns = tokenize_columns or [sen_column]

        pipeline = None
        if convert_numerical:

            def preprocess_field(s: Union[str, int]) -> Union[str, int]:
                return int(s) if (isinstance(s, str) and s.isdigit()) else s

            pipeline = Pipeline(convert_token=preprocess_field)

        fields = []

        for column in header:
            if column in tokenize_columns:
                field = Field(batch_first=True, include_lengths=True, lower=to_lower)
                if tokenizer is not None:
                    attach_tokenizer(field, tokenizer)
            else:
                field = RawField(preprocessing=pipeline)
                field.is_target = False

            fields.append((column, field))

        return fields

    @staticmethod
    def create_examples(
        raw_corpus: List[List[str]], fields: List[Tuple[str, Field]]
    ) -> List[Example]:
        examples = [Example.fromlist(line, fields) for line in raw_corpus]

        return examples

    def slice(self, sen_ids: List[int]) -> "Corpus":
        """Returns a new Corpus only containing examples from sen_ids.

        Parameters
        ----------
        sen_ids : List[int]
            List of sentence indices based on which the examples in the
            current Corpus will be filtered. These indices refer to the
            sen_idx in the original corpus; the newly sliced corpus
            will retain the original sen_idx of an Example item.

        Returns
        -------
        subcorpus : Corpus
            A new Corpus instance containing the filtered list of
            Examples.
        """
        examples = [ex for ex in self.examples if ex.sen_idx in sen_ids]

        subcorpus = Corpus(
            examples,
            self.fields,
            sen_column=self.sen_column,
            labels_column=self.labels_column,
        )

        return subcorpus

    def _attach_sen_ids(self):
        """ Adds a sentence index field to the Corpus. """
        self.fields["sen_idx"] = RawField()
        self.fields["sen_idx"].is_target = False

        for sen_idx, item in enumerate(self.examples):
            setattr(item, "sen_idx", sen_idx)

    def _create_pos_tags(self):
        """ Attaches nltk POS tags to the corpus for each sentence. """
        import nltk

        nltk.download("averaged_perceptron_tagger")

        self.fields["pos_tags"] = RawField()
        self.fields["pos_tags"].is_target = False

        print("Tagging corpus...")
        for item in self.examples:
            sen = getattr(item, self.sen_column)
            setattr(item, "pos_tags", [t[1] for t in nltk.pos_tag(sen)])


def attach_tokenizer(field: Field, tokenizer: PreTrainedTokenizer) -> None:
    """Creates a tokenizer that is attached to a Corpus Field.

    Parameters
    ----------
    field : Field
        Field to which the vocabulary will be attached
    tokenizer : PreTrainedTokenizer
        Tokenizer that will convert tokens to their index.
    """

    def preprocess(value: Union[str, List[str]]) -> List[str]:
        """We only perform the splitting as a preprocessing step.

        This allows us to still have access to the original tokens,
        including those that will be mapped to <unk> later.
        """
        if isinstance(value, list):
            value = " ".join(value)

        return [tokenizer.convert_ids_to_tokens(t) for t in tokenizer.encode(value)]

    field.preprocessing = preprocess
    field.pad_token = tokenizer.pad_token
    field.vocab = tokenizer
    field.vocab.stoi = tokenizer.vocab



"""The library currently supports the following tasks.

New tasks can be added by inheriting from :class:`SyntaxEvalTask`.
"""

from .lakretz import LakretzTask
from .linzen import LinzenTask
from .marvin import MarvinTask
from .warstadt import WarstadtTask
from .winobias import WinobiasTask
